from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_2 = StrOption(name="condition", doc="No change", default="no", properties=frozenset({"mandatory", "normal"}))
option_3 = StrOption(name="condition2", doc="No change", default="no", properties=frozenset({"mandatory", "normal"}))
option_4 = StrOption(name="test_variable", doc="No change", default="no", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal", Calculation(func.calc_value, Params(ParamValue('disabled'), kwargs={'condition': ParamOption(option_2, notraisepropertyerror=True), 'expected': ParamValue("yes"), 'reverse_condition': ParamValue(True)}), func.calc_value_property_help), Calculation(func.calc_value, Params(ParamValue('disabled'), kwargs={'condition': ParamOption(option_2, notraisepropertyerror=True), 'expected': ParamValue("maybe"), 'reverse_condition': ParamValue(True)}), func.calc_value_property_help), Calculation(func.calc_value, Params(ParamValue('disabled'), kwargs={'condition': ParamOption(option_3, notraisepropertyerror=True), 'expected': ParamValue("yes")}), func.calc_value_property_help), Calculation(func.calc_value, Params(ParamValue('disabled'), kwargs={'condition': ParamOption(option_3, notraisepropertyerror=True), 'expected': ParamValue("maybe")}), func.calc_value_property_help)}))
optiondescription_1 = OptionDescription(name="general", doc="general", children=[option_2, option_3, option_4], properties=frozenset({"normal"}))
optiondescription_24 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_1], properties=frozenset({"normal"}))
option_13 = FilenameOption(name="name", doc="name", default="/tmp/file1")
option_14 = StrOption(name="source", doc="source", default="file1")
option_12 = BoolOption(name="activate", doc="activate", default=Calculation(func.calc_value, Params((ParamValue(False)), kwargs={'default': ParamValue(True), 'condition_0': ParamOption(option_2, notraisepropertyerror=True), 'expected_0': ParamValue(["yes", "maybe"]), 'reverse_condition_0': ParamValue(True), 'condition_operator': ParamValue("OR"), 'condition_1': ParamOption(option_3, notraisepropertyerror=True), 'expected_1': ParamValue(["yes", "maybe"])})))
optiondescription_11 = OptionDescription(name="file1", doc="file1", children=[option_13, option_14, option_12])
optiondescription_10 = OptionDescription(name="files", doc="files", children=[optiondescription_11])
option_9 = BoolOption(name="activate", doc="activate", default=True)
option_15 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_26 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_10, option_9, option_15])
optiondescription_26.impl_set_information('type', "service")
optiondescription_25 = OptionDescription(name="services", doc="services", children=[optiondescription_26], properties=frozenset({"hidden"}))
optiondescription_23 = OptionDescription(name="1", doc="1", children=[optiondescription_24, optiondescription_25])
option_6 = StrOption(name="condition", doc="No change", default="no", properties=frozenset({"mandatory", "normal"}))
option_7 = StrOption(name="condition2", doc="No change", default="no", properties=frozenset({"mandatory", "normal"}))
option_8 = StrOption(name="test_variable", doc="No change", default="no", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal", Calculation(func.calc_value, Params(ParamValue('disabled'), kwargs={'condition': ParamOption(option_6, notraisepropertyerror=True), 'expected': ParamValue("yes"), 'reverse_condition': ParamValue(True)}), func.calc_value_property_help), Calculation(func.calc_value, Params(ParamValue('disabled'), kwargs={'condition': ParamOption(option_6, notraisepropertyerror=True), 'expected': ParamValue("maybe"), 'reverse_condition': ParamValue(True)}), func.calc_value_property_help), Calculation(func.calc_value, Params(ParamValue('disabled'), kwargs={'condition': ParamOption(option_7, notraisepropertyerror=True), 'expected': ParamValue("yes")}), func.calc_value_property_help), Calculation(func.calc_value, Params(ParamValue('disabled'), kwargs={'condition': ParamOption(option_7, notraisepropertyerror=True), 'expected': ParamValue("maybe")}), func.calc_value_property_help)}))
optiondescription_5 = OptionDescription(name="general", doc="general", children=[option_6, option_7, option_8], properties=frozenset({"normal"}))
optiondescription_28 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_5], properties=frozenset({"normal"}))
option_20 = FilenameOption(name="name", doc="name", default="/tmp/file1")
option_21 = StrOption(name="source", doc="source", default="file1")
option_19 = BoolOption(name="activate", doc="activate", default=Calculation(func.calc_value, Params((ParamValue(False)), kwargs={'default': ParamValue(True), 'condition_0': ParamOption(option_6, notraisepropertyerror=True), 'expected_0': ParamValue(["yes", "maybe"]), 'reverse_condition_0': ParamValue(True), 'condition_operator': ParamValue("OR"), 'condition_1': ParamOption(option_7, notraisepropertyerror=True), 'expected_1': ParamValue(["yes", "maybe"])})))
optiondescription_18 = OptionDescription(name="file1", doc="file1", children=[option_20, option_21, option_19])
optiondescription_17 = OptionDescription(name="files", doc="files", children=[optiondescription_18])
option_16 = BoolOption(name="activate", doc="activate", default=True)
option_22 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_30 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_17, option_16, option_22])
optiondescription_30.impl_set_information('type', "service")
optiondescription_29 = OptionDescription(name="services", doc="services", children=[optiondescription_30], properties=frozenset({"hidden"}))
optiondescription_27 = OptionDescription(name="2", doc="2", children=[optiondescription_28, optiondescription_29])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_23, optiondescription_27])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['frozen_rougail.dyn{{ identifier }}.vardyn'] = "{% if _.var2 == \"no\" %}\nvar2 is no\n{% endif %}\n"
dict_env['hidden_rougail.dyn{{ identifier }}'] = "{% if _.var2 == \"no\" %}\nvar2 is no\n{% endif %}\n"
option_2 = StrOption(name="var", doc="A suffix variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_3 = StrOption(name="var2", doc="a second variable", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
option_5 = StrOption(name="vardyn", doc="a dynamic variable", default="val", properties=frozenset({"force_default_on_freeze", "mandatory", "standard", Calculation(func['jinja_to_property'], Params((ParamValue("frozen")), kwargs={'__internal_jinja': ParamValue("frozen_rougail.dyn{{ identifier }}.vardyn"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_calc2/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("frozen"), '__internal_variable': ParamValue("rougail.dyn{{ identifier }}.vardyn"), 'when': ParamValue(True), 'inverse': ParamValue(False), '_.var2': ParamOption(option_3, notraisepropertyerror=True)}), help_function=func['jinja_to_property_help'])}), informations={'type': 'string'})
optiondescription_4 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="A dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_2)))), children=[option_5], properties=frozenset({"standard", Calculation(func['jinja_to_property'], Params((ParamValue("hidden")), kwargs={'__internal_jinja': ParamValue("hidden_rougail.dyn{{ identifier }}"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_calc2/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("hidden"), '__internal_variable': ParamValue("rougail.dyn{{ identifier }}"), 'when': ParamValue(True), 'inverse': ParamValue(False), '_.var2': ParamOption(option_3, notraisepropertyerror=True)}), help_function=func['jinja_to_property_help'])}), informations={'dynamic_variable': 'rougail.var'})
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_2, option_3, optiondescription_4], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



"""Redefine Tiramisu object

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Any

try:
    from tiramisu5 import DynOptionDescription, calc_value
except ModuleNotFoundError:
    from tiramisu import DynOptionDescription, calc_value
from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import (
    spec_from_loader as _spec_from_loader,
    module_from_spec as _module_from_spec,
)
from jinja2 import StrictUndefined, DictLoader
from jinja2.sandbox import SandboxedEnvironment
from rougail.object_model import CONVERT_OPTION
from rougail.error import display_xmlfiles
from tiramisu import function_waiting_for_error
from tiramisu.error import ValueWarning, ConfigError, PropertiesOptionError
from .utils import normalize_family


global func
dict_env = {}
ENV = SandboxedEnvironment(loader=DictLoader(dict_env), undefined=StrictUndefined)
func = ENV.filters
ENV.compile_templates("jinja_caches", zip=None)


class JinjaError:
    __slot__ = ("_err",)

    def __init__(self, err):
        self._err = err

    def __str__(self):
        raise self._err from self._err

    def __repr__(self):
        raise self._err from self._err

    def __eq__(self, *args, **kwargs):
        raise self._err from self._err

    def __ge__(self, *args, **kwargs):
        raise self._err from self._err

    def __gt__(self, *args, **kwargs):
        raise self._err from self._err

    def __le__(self, *args, **kwargs):
        raise self._err from self._err

    def __lt__(self, *args, **kwargs):
        raise self._err from self._err

    def __ne__(self, *args, **kwargs):
        raise self._err from self._err


def test_propertyerror(value: Any) -> bool:
    return isinstance(value, JinjaError)


ENV.tests["propertyerror"] = test_propertyerror


def load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader("func", path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith("_"):
            continue
        func[function] = getattr(func_, function)


def rougail_calc_value(*args, __default_value=None, **kwargs):
    values = calc_value(*args, **kwargs)
    if __default_value is not None and values in [None, []]:
        return __default_value
    return values


@function_waiting_for_error
def jinja_to_function(
    __internal_variable,
    __internal_attribute,
    __internal_jinja,
    __internal_type,
    __internal_multi,
    __internal_files,
    __default_value=None,
    **kwargs,
):
    global ENV, CONVERT_OPTION
    kw = {}
    for key, value in kwargs.items():
        if isinstance(value, PropertiesOptionError):
            value = JinjaError(value)
        if "." in key:
            c_kw = kw
            path, var = key.rsplit(".", 1)
            for subkey in path.split("."):
                c_kw = c_kw.setdefault(subkey, {})
            c_kw[var] = value
        else:
            if key in kw:
                raise ConfigError(
                    f'internal error, multi key for "{key}" in jinja_to_function'
                )
            kw[key] = value
    try:
        values = ENV.get_template(__internal_jinja).render(kw, **func).strip()
    except Exception as err:
        raise ConfigError(
            f'cannot calculating "{__internal_attribute}" attribute for variable "{__internal_variable}" in {display_xmlfiles(__internal_files)}: {err}'
        ) from err
    convert = CONVERT_OPTION[__internal_type].get("func", str)
    if __internal_multi:
        values = [convert(val) for val in values.split("\n") if val != ""]
        if not values and __default_value is not None:
            return __default_value
        return values
    try:
        values = convert(values)
    except Exception as err:
        raise ConfigError(
            f'cannot converting "{__internal_attribute}" attribute for variable "{__internal_variable}" in {display_xmlfiles(__internal_files)}: {err}'
        ) from err
    values = values if values != "" and values != "None" else None
    if values is None and __default_value is not None:
        return __default_value
    return values


def variable_to_property(prop, value, when, inverse):
    if isinstance(value, PropertiesOptionError):
        raise value from value
    if inverse:
        is_match = value != when
    else:
        is_match = value == when
    return prop if is_match else None


@function_waiting_for_error
def jinja_to_property(prop, when, inverse, **kwargs):
    value = func["jinja_to_function"](**kwargs)
    return func["variable_to_property"](prop, value is not None, when, inverse)


@function_waiting_for_error
def jinja_to_property_help(prop, **kwargs):
    value = func["jinja_to_function"](**kwargs)
    return (prop, f'"{prop}" ({value})')


@function_waiting_for_error
def valid_with_jinja(warnings_only=False, **kwargs):
    global ValueWarning
    value = func["jinja_to_function"](**kwargs)
    if value:
        if warnings_only:
            raise ValueWarning(value)
        else:
            raise ValueError(value)


func["calc_value"] = rougail_calc_value
func["jinja_to_function"] = jinja_to_function
func["jinja_to_property"] = jinja_to_property
func["jinja_to_property_help"] = jinja_to_property_help
func["variable_to_property"] = variable_to_property
func["valid_with_jinja"] = valid_with_jinja


class ConvertDynOptionDescription(DynOptionDescription):
    """Identifier could be an integer, we should convert it in str
    Identifier could also contain invalid character, so we should "normalize" it
    """

    def convert_identifier_to_path(self, identifier):
        if identifier is None:
            return identifier
        if not isinstance(identifier, str):
            identifier = str(identifier)
        return normalize_family(identifier)

    def impl_getname(
        self,
        identifier=None,
    ) -> str:
        """get name"""
        name = super().impl_getname(None)
        if identifier is None:
            return name
        path_identifier = self.convert_identifier_to_path(identifier)
        if "{{ identifier }}" in name:
            return name.replace("{{ identifier }}", path_identifier)
        return name + path_identifier

    def impl_get_display_name(
        self,
        subconfig,
        with_quote: bool = False,
    ) -> str:
        display = super().impl_get_display_name(subconfig, with_quote=with_quote)
        if "{{ identifier }}" in display:
            return display.replace(
                "{{ identifier }}",
                self.convert_identifier_to_path(self.get_identifiers(subconfig)[-1]),
            )
        return display



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_2 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"mandatory", "normal"}))
optiondescription_1 = OptionDescription(name="general", doc="general", children=[option_2], properties=frozenset({"normal"}))
option_5 = StrOption(name="leader", doc="leader", multi=True, properties=frozenset({"normal"}))
option_6 = StrOption(name="follower1", doc="follower1", multi=True, default=Calculation(func.calc_val, Params((), kwargs={'valeur': ParamValue("valfill")})), properties=frozenset({"normal"}))
option_7 = StrOption(name="follower2", doc="follower2", multi=submulti, default=[Calculation(func.calc_val, Params((ParamOption(option_6))))], properties=frozenset({"normal"}))
optiondescription_4 = Leadership(name="leader", doc="leader", children=[option_5, option_6, option_7], properties=frozenset({"normal"}))
optiondescription_3 = OptionDescription(name="general1", doc="general1", children=[optiondescription_4], properties=frozenset({"normal"}))
optiondescription_16 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_1, optiondescription_3], properties=frozenset({"normal"}))
optiondescription_15 = OptionDescription(name="1", doc="1", children=[optiondescription_16])
option_9 = StrOption(name="mode_conteneur_actif", doc="No change", default="non", properties=frozenset({"mandatory", "normal"}))
optiondescription_8 = OptionDescription(name="general", doc="general", children=[option_9], properties=frozenset({"normal"}))
option_12 = StrOption(name="leader", doc="leader", multi=True, properties=frozenset({"normal"}))
option_13 = StrOption(name="follower1", doc="follower1", multi=True, default=Calculation(func.calc_val, Params((), kwargs={'valeur': ParamValue("valfill")})), properties=frozenset({"normal"}))
option_14 = StrOption(name="follower2", doc="follower2", multi=submulti, default=[Calculation(func.calc_val, Params((ParamOption(option_13))))], properties=frozenset({"normal"}))
optiondescription_11 = Leadership(name="leader", doc="leader", children=[option_12, option_13, option_14], properties=frozenset({"normal"}))
optiondescription_10 = OptionDescription(name="general1", doc="general1", children=[optiondescription_11], properties=frozenset({"normal"}))
optiondescription_18 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_8, optiondescription_10], properties=frozenset({"normal"}))
optiondescription_17 = OptionDescription(name="2", doc="2", children=[optiondescription_18])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_15, optiondescription_17])



from os.path import isfile, join, isdir
from pytest import fixture
from os import listdir, mkdir, environ
from json import dump, load, dumps, loads
from pathlib import Path

environ['TIRAMISU_LOCALE'] = 'en'

from .custom import CustomOption
from tiramisu import Config
from tiramisu.error import PropertiesOptionError


dico_dirs = 'tests/dictionaries'


test_ok = set()

for test in listdir(dico_dirs):
    if isdir(join(dico_dirs, test)):
        if isdir(join(dico_dirs, test, 'tiramisu')):
            test_ok.add(test)

debug = False
#debug = True
excludes = set([])
#excludes = set([
#    '80leadership_subfamily',
#    '80valid_enum_variables',
#])

# excludes = set(['60_5family_dynamic_variable_outside_sub_suffix'])
test_ok -= excludes
# test_ok = ['04_1default_calculation_hidden']


test_ok = list(test_ok)
test_ok.sort()
no_test_base_multi = False
#no_test_base_multi = True
#no_test_base = True
no_test_base = False
#no_test_multi = True
no_test_multi = False


@fixture(scope="module", params=test_ok)
def test_dir(request):
    return request.param


def option_value(parent, key_is_option=False):
    for option, value in parent.items():
        if option.isoptiondescription():
            if not key_is_option and option.isleadership():
                ret = []
                for idx, datas in enumerate(option_value(value, key_is_option=True)):
                    sub_option, sub_value = datas
                    if not idx:
                        sub_option = sub_option.path()
                        key = sub_option
                        for val in sub_value:
                            ret.append({sub_option: val})
                    else:
                        index = sub_option.index()
                        sub_option = sub_option.path()
                        ret[index][sub_option] = sub_value
                yield key, ret
            else:
                yield from option_value(value, key_is_option)
        elif key_is_option:
            yield option, value
        else:
            yield option.path(), value


def launch_flattener(test_dir,
                     filename,
                     ):
    makedict_dir = join(test_dir, 'makedict')
    makedict_file = join(makedict_dir, 'base.json')
    makedict_before = join(makedict_dir, 'before.json')
    makedict_after = join(makedict_dir, 'after.json')
    informations_file = join(test_dir, 'informations.json')
    mandatory_file = Path(makedict_dir) / 'mandatory.json'

    modulepath = join(test_dir, 'tiramisu', filename + '.py')
    with open(modulepath) as fh:
        optiondescription = {}
        exec(fh.read(), {'CustomOption': CustomOption}, optiondescription)  # pylint: disable=W0122
    config = Config(optiondescription["option_0"])
    # change default rights
    ro_origin = config.property.default('read_only', 'append')
    ro_append = frozenset(ro_origin - {'force_store_value'})
    rw_origin = config.property.default('read_write', 'append')
    rw_append = frozenset(rw_origin - {'force_store_value'})
    config.property.setdefault(ro_append, 'read_only', 'append')
    config.property.setdefault(rw_append, 'read_write', 'append')

    config.information.set('test_information', 'value')
    config.property.read_only()
    config.property.remove('mandatory')
    config.information.set('info', 'value')
    if isfile(informations_file):
        with open(informations_file) as informations:
            for key, value in load(informations).items():
                if filename == 'base':
                    config.option(key).information.set('test_information', value)
                elif filename == 'no_namespace':
                    config.option(key.split('.', 1)[-1]).information.set('test_information', value)
                else:
                    for root in ['1', '2']:
                        config.option(f'{root}.{key}').information.set('test_information', value)
    #
    config_dict = dict(option_value(config.value.get()))
    if filename == 'base':
        if not isdir(makedict_dir):
            mkdir(makedict_dir)
        if not isfile(makedict_file) or debug:
            with open(makedict_file, 'w') as fh:
                dump(config_dict, fh, indent=4)
                fh.write('\n')
    elif filename == 'no_namespace':
        config_dict = {f'rougail.{path}': value for path, value in config_dict.items()}
    else:
        config_dict_prefix = {'1': {}, '2': {}}
        for key, value in config_dict.items():
            prefix, path = key.split('.', 1)
            if value and isinstance(value, list) and isinstance(value[0], dict):
                new_value = []
                for dct in value:
                    new_dct = {}
                    for k, v in dct.items():
                        k = k.split('.', 1)[-1]
                        new_dct[k] = v
                    new_value.append(new_dct)
                value = new_value
            config_dict_prefix[prefix][path] = value
        assert loads(dumps(config_dict_prefix['1'])) == loads(dumps(config_dict_prefix['2']))
        config_dict = config_dict_prefix['1']
    if not isfile(makedict_file):
        raise Exception('dict is not empty')
    with open(makedict_file, 'r') as fh:
        assert load(fh) == loads(dumps(config_dict)), f"error in file {makedict_file}"
    #
    value_owner(makedict_before, config, filename)
    # deploy
    ro = config.property.default('read_only', 'append')
    ro = frozenset(list(ro) + ['force_store_value'])
    config.property.setdefault(ro, 'read_only', 'append')
    rw = config.property.default('read_write', 'append')
    rw = frozenset(list(rw) + ['force_store_value'])
    config.property.setdefault(rw, 'read_write', 'append')
    config.property.add('force_store_value')
    #
    value_owner(makedict_after, config, filename)
    #
    mandatory(mandatory_file, config.value.mandatory(), filename)


def value_owner(makedict_value_owner, config, filename):
    ret = {}
    for key, value in option_value(config.value.get(), True):
        path = key.path()
        if not key.issymlinkoption() and key.isfollower():
            if path in ret:
                continue
            ret[path] = {'owner': [],
                         'value': [],
                         }
            for idx in range(0, key.value.len()):
                try:
                    option = config.option(path, idx)
                    ret[path]['value'].append(option.value.get())
                    ret[path]['owner'].append(option.owner.get())
                except PropertiesOptionError as err:
                    ret[path]['value'].append(str(err))
                    ret[path]['owner'].append('error')
        else:
            owner = key.owner.get()
            ret[path] = {'owner': owner,
                         'value': value,
                         }
    if filename == 'base':
        if not isfile(makedict_value_owner) or debug:
            with open(makedict_value_owner, 'w') as fh:
                dump(ret, fh, indent=4)
                fh.write('\n')
    elif filename == 'no_namespace':
        ret = {f'rougail.{path}': value for path, value in ret.items()}
    else:
        ret_prefix = {'1': {}, '2': {}}
        for key, value in ret.items():
            prefix, path = key.split('.', 1)
            ret_prefix[prefix][path] = value
        assert loads(dumps(ret_prefix['1'])) == loads(dumps(ret_prefix['2']))
        ret = ret_prefix['1']
    with open(makedict_value_owner, 'r') as fh:
        assert load(fh) == loads(dumps(ret)), f"error in file {makedict_value_owner}"


def mandatory(mandatory_file, mandatories, filename):
    ret = [opt.path() for opt in mandatories]
    if not mandatory_file.is_file():
        with mandatory_file.open('w') as fh:
            dump(ret, fh)
    if filename == 'no_namespace':
        ret = [f'rougail.{path}' for path in ret]
    elif filename == 'multi':
        ret_prefix = {'1': [], '2': []}
        for key in ret:
            prefix, path = key.split('.', 1)
            ret_prefix[prefix].append(path)
        assert ret_prefix['1'] == ret_prefix['2']
        ret = ret_prefix['1']
    with mandatory_file.open() as fh:
        assert ret == load(fh), f"error in file {mandatory_file}"


def test_dictionary(test_dir):
    if no_test_base:
        print('FIXME')
        return
    test_dir = join(dico_dirs, test_dir)
    launch_flattener(test_dir, 'base')


def test_dictionary_no_namespace(test_dir):
    test_dir = join(dico_dirs, test_dir)
    if isfile(join(test_dir, 'force_namespace')):
        return
    launch_flattener(test_dir, 'no_namespace')


def test_dictionary_multi(test_dir):
    if no_test_base_multi:
        print('FIXME')
        return
    test_dir = join(dico_dirs, test_dir)
    launch_flattener(test_dir, 'multi')



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['default_var2'] = "{{ _.var1 }}\n"
option_1 = StrOption(name="var1", doc="a first variable", default="no", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_2 = StrOption(name="var2", doc="a second variable", multi=True, default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_var2"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(True), '__internal_files': ParamValue(['tests/dictionaries/00_2default_calculated/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("var2"), '_.var1': ParamOption(option_1, notraisepropertyerror=True)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1, option_2])



from pathlib import Path
from pytest import fixture, raises
from shutil import rmtree
from rougail import Rougail, RougailConfig

doc = Path('doc')
tests = Path('tests')
tmp = tests / 'tmp'
if tmp.is_dir():
    rmtree(tmp)
tmp.mkdir()
RougailConfig['main_dictionaries'] = [str(tmp)]
RougailConfig['functions_files'] = [str(tests / 'eosfunc' / 'test.py')]


scripts = {}
for filename in doc.glob('*/*.md'):
    filename = str(filename)
    with open(filename) as fh:
        yaml = False
        redefine = False
        new_script = ''
        for line in fh.readlines():
            if new_script and line == '```\n':
                scripts.setdefault(filename, [])
                if redefine:
                    scripts[filename][-1].append(new_script)
                else:
                    scripts[filename].append([new_script])
                yaml = False
                redefine = False
            if yaml:
                if line.strip() == 'redefine: true':
                    redefine = True
                new_script += line
            if line == '```yml\n':
                yaml = True
                new_script = ''
scripts_list = [(filename, script) for filename, scripts_ in scripts.items() for script in scripts_]


@fixture(scope="module", params=scripts_list)
def test_dir(request):
    return request.param


def test_scripts(test_dir):
    if tmp.is_dir():
        rmtree(tmp)
    tmp.mkdir()
    for idx, content in enumerate(test_dir[1]):
        if not content.startswith('---'):
            raise Exception(f'not a valid template in {test_dir[0]}')
        filename = tmp / f'0{idx}-base.yml'
        with open(filename, 'w') as fh:
            fh.write(content)
    rougail = Rougail()
    try:
        config = rougail.get_config()
        config.value.dict()
    except Exception as err:
        #rmtree(tmp)
        raise Exception(f'error in {test_dir[0]}: {err}') from err
    rmtree(tmp)



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['disabled_rougail.family'] = "true\n"
option_3 = StrOption(name="var1", doc="var1", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_2 = OptionDescription(name="family", doc="family", children=[option_3], properties=frozenset({"basic", Calculation(func['jinja_to_property'], Params((ParamValue("disabled")), kwargs={'__internal_jinja': ParamValue("disabled_rougail.family"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/16_2family_redefine_calculation/dictionaries/rougail/01-base.yml']), '__internal_attribute': ParamValue("disabled"), '__internal_variable': ParamValue("rougail.family"), 'when': ParamValue(True), 'inverse': ParamValue(False)}), help_function=func['jinja_to_property_help'])}))
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[optiondescription_2], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['default_rougail.var2'] = "{%- for v in var %}\n{{ v }}\n{%- endfor -%}"
option_2 = StrOption(name="var", doc="a suffix variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_4 = StrOption(name="var", doc="a variable inside a dynamic family", default=Calculation(func['calc_value'], Params((ParamIdentifier()))), properties=frozenset({"standard"}), informations={'type': 'string'})
optiondescription_3 = ConvertDynOptionDescription(name="my_dyn_family_{{ identifier }}", doc="a dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_2)))), children=[option_4], properties=frozenset({"standard"}), informations={'dynamic_variable': 'rougail.var'})
option_5 = StrOption(name="var2", doc="a variable", multi=True, default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_rougail.var2"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(True), '__internal_files': ParamValue(['tests/dictionaries/60_5family_dynamic_variable_outside_jinja/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("rougail.var2"), 'var': ParamOption(option_4)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_1 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_2, optiondescription_3, option_5], properties=frozenset({"standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



"""Rougail's tools

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import List, Union
from unicodedata import normalize, combining
import re
from itertools import chain

from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader, module_from_spec

from jinja2 import DictLoader, TemplateSyntaxError
from jinja2.sandbox import SandboxedEnvironment
from jinja2.parser import Parser
from jinja2.nodes import Name, Getattr

from tiramisu.config import get_common_path

from .i18n import _
from .error import DictConsistencyError

NAME_REGEXP = re.compile(r"^[a-z0-9_]*$")


def valid_variable_family_name(
    name: str,
    xmlfiles: List[str],
) -> None:
    match = NAME_REGEXP.search(name)
    if not match:
        msg = _(
            'invalid variable or family name "{0}" must only contains lowercase ascii character, number or _'
        ).format(name)
        raise DictConsistencyError(msg, 76, xmlfiles)


def normalize_family(family_name: str) -> str:
    """replace space, accent, uppercase, ... by valid character"""
    if not family_name:
        return
    family_name = family_name.lower()
    family_name = family_name.replace("-", "_").replace(" ", "_").replace(".", "_")
    nfkd_form = normalize("NFKD", family_name)
    family_name = "".join([c for c in nfkd_form if not combining(c)])
    return family_name.lower()


def load_modules(name, module) -> List[str]:
    """list all functions in a module"""
    loader = SourceFileLoader(name, module)
    spec = spec_from_loader(loader.name, loader)
    eosfunc = module_from_spec(spec)
    loader.exec_module(eosfunc)
    return eosfunc


def get_realpath(
    path: str,
    path_prefix: str,
) -> str:
    if path_prefix:
        return f"{path_prefix}.{path}"
    return path


def get_jinja_variable_to_param(
    current_path: str,
    jinja_text,
    objectspace,
    xmlfiles,
    functions,
    path_prefix,
    version,
    namespace,
):
    try:
        env = SandboxedEnvironment(loader=DictLoader({"tmpl": jinja_text}))
        env.filters = functions
        parsed_content = Parser(env, jinja_text, "", "").parse()

        def recurse_getattr(g: Getattr):
            if isinstance(g.node, Getattr):
                return recurse_getattr(g.node) + "." + g.attr
            return g.node.name + "." + g.attr

        variables = set()
        if objectspace.namespace is None:
            for n in parsed_content.find_all(Name):
                variables.add(n.name)
        for g in parsed_content.find_all(Getattr):
            variables.add(recurse_getattr(g))
    except TemplateSyntaxError as err:
        msg = _('error in jinja "{0}" for the variable "{1}": {2}').format(
            jinja_text, current_path, err
        )
        raise DictConsistencyError(msg, 39, xmlfiles) from err
    variables = list(variables)
    variables.sort(reverse=True)
    founded_variables = {}
    unknown_variables = []
    for variable_path in variables:
        variable, identifier = objectspace.paths.get_with_dynamic(
            variable_path,
            path_prefix,
            current_path,
            version,
            namespace,
            xmlfiles,
        )
        if variable and variable.path in objectspace.variables:
            founded_variables[variable_path] = (identifier, variable)
        else:
            sub_family = variable_path + "."
            for founded_variable in chain(founded_variables, unknown_variables):
                if founded_variable.startswith(sub_family):
                    break
            else:
                unknown_variables.append(variable_path)

    for variable_path in unknown_variables:
        for v in founded_variables:
            if get_common_path(v, variable_path) == v:
                break
        else:
            root_path = None
            vpath = variable_path
            while "." in vpath:
                vpath = vpath.rsplit(".", 1)[0]
                variable, identifier = objectspace.paths.get_with_dynamic(
                    vpath,
                    path_prefix,
                    current_path,
                    version,
                    namespace,
                    xmlfiles,
                )
                if variable and variable.path in objectspace.families:
                    root_path = vpath
                    break
            if root_path:
                yield {}, None, root_path
    for variable_path, data in founded_variables.items():
        yield data[1], data[0], variable_path



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries_old/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_3 = StrOption(name="mode_conteneur_actif", doc="Description", default="non", properties=frozenset({"mandatory", "normal"}))
optiondescription_2 = OptionDescription(name="general", doc="general", children=[option_3], properties=frozenset({"normal"}))
optiondescription_1 = OptionDescription(name="rougail", doc="rougail", children=[optiondescription_2], properties=frozenset({"normal"}))
option_8 = FilenameOption(name="name", doc="name", default="/etc/file")
option_9 = FilenameOption(name="source", doc="source", default="/etc/file")
option_10 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_7 = OptionDescription(name="file", doc="file", children=[option_8, option_9, option_10])
option_12 = FilenameOption(name="name", doc="name", default="/etc/file2")
option_13 = FilenameOption(name="source", doc="source", default="/etc/file2")
option_14 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_11 = OptionDescription(name="file2", doc="file2", children=[option_12, option_13, option_14])
optiondescription_11.impl_set_information('engine', "jinja")
option_16 = FilenameOption(name="name", doc="name", default="/etc/file3")
option_17 = FilenameOption(name="source", doc="source", default="/etc/file3")
option_18 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_15 = OptionDescription(name="file3", doc="file3", children=[option_16, option_17, option_18])
optiondescription_6 = OptionDescription(name="files", doc="files", children=[optiondescription_7, optiondescription_11, optiondescription_15])
option_19 = BoolOption(name="activate", doc="activate", default=True)
option_20 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_5 = OptionDescription(name="test_service", doc="test_service", children=[optiondescription_6, option_19, option_20])
optiondescription_5.impl_set_information('type', "service")
optiondescription_4 = OptionDescription(name="services", doc="services", children=[optiondescription_5], properties=frozenset({"hidden", "normal"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_4])



def trange(start, stop):
    return list(range(start, stop))


def calc_val(*args, **kwargs):
    if len(args) > 0:
        return args[0]


def concat(*args, **kwargs):
    pass


def calc_multi_condition(*args, **kwargs):
    pass


def get_zone_name_bridge(*args, **kwargs):
    pass


def get_devices(*args, **kwargs):
    pass


def get_mount_point_device(*args, **kwargs):
    pass


def valid_differ(*args, **kwargs):
    pass


def valid_ipnetmask(*args, **kwargs):
    pass


def valid_enum(*args, **kwargs):
    pass


def valid_lower(*args, **kwargs):
    pass


def list_files(*args, **kwargs):
    # FIXME ?
    return kwargs['default']


def calc_multi_val(*args, **kwargs):
    return args[0]


def cdrom_minormajor(*args, **kwargs):
    pass


def device_type(*args, **kwargs):
    pass


def calc_list(*args, **kwargs):
    return list(args)


def test_index(index):
    return index


def return_no():
    return 'no'


def return_yes():
    return 'yes'



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
from rougail.tiramisu import ConvertDynOptionDescription
option_2 = StrOption(name="varname", doc="No change", multi=True, default=['val1', 'val2'], default_multi="val1", properties=frozenset({"mandatory", "normal"}))
optiondescription_1 = OptionDescription(name="general", doc="general", children=[option_2], properties=frozenset({"normal"}))
option_4 = StrOption(name="vardyn", doc="No change", default=Calculation(func.calc_val, Params((ParamSuffix()))), properties=frozenset({"normal"}))
optiondescription_3 = ConvertDynOptionDescription(name="dyn", doc="dyn", suffixes=Calculation(func.calc_value, Params((ParamOption(option_2, notraisepropertyerror=True)))), children=[option_4], properties=frozenset({"normal"}))
option_6 = StrOption(name="newvar", doc="No change", properties=frozenset({"normal"}))
optiondescription_5 = OptionDescription(name="new", doc="new", children=[option_6], properties=frozenset({"normal"}))
optiondescription_14 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_1, optiondescription_3, optiondescription_5], properties=frozenset({"normal"}))
optiondescription_13 = OptionDescription(name="1", doc="1", children=[optiondescription_14])
option_8 = StrOption(name="varname", doc="No change", multi=True, default=['val1', 'val2'], default_multi="val1", properties=frozenset({"mandatory", "normal"}))
optiondescription_7 = OptionDescription(name="general", doc="general", children=[option_8], properties=frozenset({"normal"}))
option_10 = StrOption(name="vardyn", doc="No change", default=Calculation(func.calc_val, Params((ParamSuffix()))), properties=frozenset({"normal"}))
optiondescription_9 = ConvertDynOptionDescription(name="dyn", doc="dyn", suffixes=Calculation(func.calc_value, Params((ParamOption(option_8, notraisepropertyerror=True)))), children=[option_10], properties=frozenset({"normal"}))
option_12 = StrOption(name="newvar", doc="No change", properties=frozenset({"normal"}))
optiondescription_11 = OptionDescription(name="new", doc="new", children=[option_12], properties=frozenset({"normal"}))
optiondescription_16 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_7, optiondescription_9, optiondescription_11], properties=frozenset({"normal"}))
optiondescription_15 = OptionDescription(name="2", doc="2", children=[optiondescription_16])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_13, optiondescription_15])



"""Annotate value

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from rougail.annotator.variable import Walk

from rougail.i18n import _
from rougail.error import DictConsistencyError
from rougail.object_model import Calculation


class Annotator(Walk):  # pylint: disable=R0903
    """Annotate value"""

    level = 70

    def __init__(
        self,
        objectspace,
        *args,
    ) -> None:
        if not objectspace.paths:
            return
        self.objectspace = objectspace
        self.convert_value()
        self.valid_choices()

    def convert_value(self) -> None:
        """convert value"""
        for variable in self.get_variables():
            if variable.type == "symlink":
                continue
            if variable.version != "1.0" and variable.type == "port":
                self._convert_port(variable)
            self._convert_value(variable)

    def _convert_value(
        self,
        variable: dict,
    ) -> None:
        multi = self.objectspace.multis.get(variable.path, False)
        # a boolean must have value, the default value is "True"
        if variable.type == "boolean" and multi is False and variable.default is None:
            variable.default = True

        if variable.default is None or isinstance(variable.default, Calculation):
            return

        if isinstance(variable.default, list):
            if not multi:
                msg = f'The variable "{variable.path}" with a list as default value must have "multi" attribute'
                raise DictConsistencyError(msg, 68, variable.xmlfiles)
            if variable.path in self.objectspace.followers and multi != "submulti":
                msg = _(
                    'the follower "{0}" without multi attribute can only have one value'
                ).format(variable.name)
                raise DictConsistencyError(msg, 87, variable.xmlfiles)
            if not variable.default:
                variable.default = None
            else:
                if variable.path not in self.objectspace.leaders:
                    if multi == "submulti":
                        self.objectspace.default_multi[variable.path] = variable.default
                        variable.default = None
                    else:
                        self.objectspace.default_multi[variable.path] = (
                            variable.default[0]
                        )
        elif variable.multi:
            msg = _(
                'the variable "{0}" is multi but has a non list default value'
            ).format(variable.name)
            raise DictConsistencyError(msg, 12, variable.xmlfiles)
        elif variable.path in self.objectspace.followers:
            self.objectspace.default_multi[variable.path] = variable.default
            variable.default = None

    def _convert_port(self, variable) -> None:
        if variable.multi is False and isinstance(variable.default, int):
            variable.default = str(variable.default)
        elif variable.multi is True and isinstance(variable.default, list):
            for idx, value in enumerate(variable.default):
                if isinstance(value, int):
                    variable.default[idx] = str(value)

    def valid_choices(self) -> None:
        """A variable with type "Choice" that is not mandatory must has "nil" value"""
        for variable in self.get_variables():
            if variable.type != "choice":
                continue
            if isinstance(variable.choices, Calculation):
                continue
            if variable.choices is None:
                msg = f'the variable "{variable.path}" is a "choice" variable but don\'t have any choice'
                raise DictConsistencyError(msg, 19, variable.xmlfiles)
            if not variable.mandatory and not variable.multi:
                self.add_choice_nil(variable)

    def add_choice_nil(self, variable) -> None:
        """A variable with type "Choice" that is not mandatory must has "nil" value"""
        for choice in variable.choices:
            if choice is None:
                return
        variable.choices.append(None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    author='Silique',
    author_email='egarette@silique.fr',
    name='rougail',
    version='0.1',
    description='Configuration templating engine',
    url='https://forge.cadoles.com/Infra/rougail',
    packages=find_packages('src'),
    package_dir={"": "src"},
    include_package_data=True,
    package_data={"": ["data/*.dtd"]}
)



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['default_variable'] = "{{ param1 }}_{{ param2 }}_{{ param3 }}_{{ param4 }}\n"
option_1 = StrOption(name="variable", doc="a variable", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_variable"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/00_9default_calculation/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("variable"), 'param1': ParamValue("string"), 'param2': ParamValue(1), 'param3': ParamValue(True), 'param4': ParamValue(None)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[option_1])



"""Update Rougail structure file to new version

Cadoles (http://www.cadoles.com)
Copyright (C) 2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from os import listdir
from os.path import basename, isdir, isfile, join
from typing import Any, List, Optional, Tuple

try:
    from lxml.etree import SubElement  # pylint: disable=E0611
    from lxml.etree import Element, XMLParser, XMLSyntaxError, parse, tostring
except ModuleNotFoundError as err:
    parse = None

# from ast import parse as ast_parse
from json import dumps
from ruamel.yaml import YAML
from pathlib import Path

from ..config import RougailConfig
from ..error import UpgradeError
from ..i18n import _
from ..object_model import CONVERT_OPTION
from ..utils import normalize_family

VERSIONS = ["0.10", "1.0", "1.1"]


def get_function_name(version):
    version = version.replace(".", "_")
    return f"update_{version}"


FUNCTION_VERSIONS = [(version, get_function_name(version)) for version in VERSIONS]


class upgrade_010_to_10:
    def __init__(
        self,
        dico: dict,
        namespace: str,
        xmlsrc: str,
    ) -> None:
        self.xmlsrc = xmlsrc
        self.paths = {"family": {}, "variable": {}, "dynamic": {}}
        self.lists = {
            "service": {},
            "ip": {},
            "certificate": {},
            "file": {},
        }
        self.flatten_paths = {"family": {}, "variable": {}}
        self.variables = self.parse_variables(dico, namespace, namespace, root=True)
        self.parse_variables_with_path()
        self.parse_services(dico)
        self.parse_constraints(dico)

    def parse_variables(
        self,
        family: dict,
        sub_path: str,
        true_sub_path: str,
        *,
        root: bool = False,
        is_dynamic: bool = False,
    ) -> dict:
        new_families = {}
        if root:
            new_families["version"] = None
        if "variables" in family:
            for subelt in family["variables"]:
                for typ, obj in subelt.items():
                    for subobj in obj:
                        local_is_dynamic = (
                            is_dynamic or subobj.get("dynamic") is not None
                        )
                        getattr(self, f"convert_{typ}")(
                            subobj,
                            new_families,
                            sub_path,
                            true_sub_path,
                            local_is_dynamic,
                        )
            family.pop("variables")
        return new_families

    def convert_family(
        self,
        family: dict,
        new_families: dict,
        sub_path: str,
        true_sub_path: str,
        is_dynamic: bool,
    ) -> None:
        # name is the key, do not let it in values
        name = family.pop("name")
        if true_sub_path:
            true_sub_path = true_sub_path + "." + name
        else:
            true_sub_path = name
        if is_dynamic:
            name += "{{ suffix }}"
        if sub_path:
            sub_path = sub_path + "." + name
        else:
            sub_path = name
        # leadership and dynamic are no more attribute, it's now a type
        for typ in ["leadership", "dynamic"]:
            if typ in family:
                value = family.pop(typ)
                if value:
                    family["type"] = typ
                if typ == "dynamic":
                    family["variable"] = self.get_variable_path(value)
        # add sub families and sub variables
        sub_families = self.parse_variables(
            family, sub_path, true_sub_path, is_dynamic=is_dynamic
        )
        for sub_name, sub_family in sub_families.copy().items():
            if sub_name not in family:
                continue
            family[f"_{sub_name}"] = family.pop(sub_name)
        # store converted family
        family.update(sub_families)
        new_families[name] = family
        self.flatten_paths["family"][name] = sub_path
        self.paths["family"][sub_path] = family

    def convert_variable(
        self,
        variable: dict,
        new_families: dict,
        sub_path: str,
        true_sub_path: str,
        is_dynamic: bool,
    ) -> dict:
        name = variable.pop("name")
        if is_dynamic:
            if true_sub_path:
                true_sub_path = true_sub_path + "." + name
            else:
                true_sub_path = name
            self.flatten_paths["variable"][name] = true_sub_path
            name += "{{ suffix }}"
        if sub_path:
            sub_path = sub_path + "." + name
        else:
            sub_path = name
        if is_dynamic:
            self.paths["dynamic"][true_sub_path] = sub_path
        new_families[name] = variable
        self.flatten_paths["variable"][name] = sub_path
        self.paths["variable"][sub_path] = variable
        if (
            "redefine" not in variable
            and "value" not in variable
            and "mandatory" not in variable
            and ("type" not in variable or variable["type"] != "boolean")
        ):
            variable["mandatory"] = False
        if "remove_condition" in variable and variable.pop("remove_condition"):
            for prop in ["hidden", "disabled", "mandatory"]:
                if prop not in variable:
                    variable[prop] = False
        if "remove_choice" in variable:
            if "choice" not in variable:
                variable["choice"] = None
            variable.pop("remove_choice")
        if "remove_check" in variable:
            variable.pop("remove_check")
            if "validators" not in variable:
                variable["validators"] = None
        if "remove_fill" in variable:
            variable.pop("remove_fill")
            variable["value"] = None
        if "auto_freeze" in variable:
            variable.pop("auto_freeze")
        if "test" in variable:
            tests = []
            for test in variable["test"].split("|"):
                if test == "":
                    tests.append(None)
                else:
                    tests.append(
                        CONVERT_OPTION.get(variable.get("type", "string"), {}).get(
                            "func", str
                        )(test)
                    )
            variable["test"] = tests
        if variable.get("mandatory", False):
            del variable["mandatory"]
        CONVERT_TYPE = {
            "filename": "unix_filename",
            "password": "secret",
        }
        if variable.get("type") in CONVERT_TYPE:
            variable["type"] = CONVERT_TYPE[variable["type"]]

    def parse_variables_with_path(self):
        for variable in self.paths["variable"].values():
            multi = variable.get("multi", False)
            if "value" in variable:
                default = variable.pop("value")
                if default is not None:
                    if not multi and len(default) == 1:
                        variable["default"] = self.get_value(default[0], multi)
                    else:
                        variable["default"] = [
                            self.get_value(value, multi) for value in default
                        ]
            if "choice" in variable:
                if not variable["choice"]:
                    variable["choices"] = variable.pop("choice")
                else:
                    variable["choices"] = [
                        self.get_value(choice, multi)
                        for choice in variable.pop("choice")
                    ]

    def parse_services(
        self,
        dico: dict,
    ) -> None:
        self.services = {}
        if "services" in dico:
            for root_services in dico["services"]:
                for services in root_services.values():
                    for service in services:
                        new_service = {}
                        for typ in ["ip", "file", "certificate"]:
                            if typ != "ip":
                                typ_plurial = typ + "s"
                            else:
                                typ_plurial = typ
                            if typ in service:
                                new_service[typ_plurial] = {}
                                for elt in service[typ]:
                                    name, new_elt = getattr(self, f"parse_{typ}")(elt)
                                    new_service[typ_plurial][name] = new_elt
                        if "override" in service:
                            if isinstance(service["override"], list):
                                new_service["override"] = service["override"][0]
                            else:
                                new_service["override"] = service["override"]
                        if "servicelist" in service:
                            self.lists["service"].setdefault(
                                service["servicelist"], []
                            ).append(new_service)
                        name = (
                            service.pop("name") + "." + service.get("type", "service")
                        )
                        self.services[name] = new_service

    def parse_ip(self, ip: dict) -> None:
        name = self.get_variable_path(ip.pop("text"))
        if "iplist" in ip:
            self.lists["ip"].setdefault(ip.pop("iplist"), []).append(ip)
        if "netmask" in ip:
            ip["netmask"] = self.get_variable_path(ip["netmask"])
        return name, ip

    def parse_file(self, file: dict) -> None:
        name = file.pop("text")
        if "file_type" in file:
            file["type"] = file.pop("file_type")
            if file["type"] == "variable":
                name = self.get_variable_path(name)
        if "variable_type" in file:
            file.pop("variable_type")
        if "filelist" in file:
            self.lists["file"].setdefault(file.pop("filelist"), []).append(file)
        for typ in ["source", "owner", "group"]:
            if f"{typ}_type" in file:
                obj_type = file.pop(f"{typ}_type")
                if obj_type == "variable" and typ in file:
                    file[typ] = {
                        "name": self.get_variable_path(file[typ]),
                        "type": "variable",
                    }
        if "variable" in file:
            file["variable"] = self.get_variable_path(file["variable"])
        return name, file

    def parse_certificate(self, certificate: dict) -> None:
        name = certificate.pop("text")
        if "variable" in certificate:
            certificate["rougail_variable"] = certificate["variable"]
        if "certificate_type" in certificate:
            if certificate.pop("certificate_type") == "variable":
                certificate["variable"] = True
                name = self.get_variable_path(name)
        if "certificatelist" in certificate:
            self.lists["certificate"].setdefault(
                certificate.pop("certificatelist"), []
            ).append(certificate)
        for typ in ["owner", "group", "server", "domain", "provider"]:
            if f"{typ}_type" in certificate:
                obj_type = certificate.pop(f"{typ}_type")
                if obj_type == "variable" and typ in certificate:
                    certificate[typ] = {
                        "name": self.get_variable_path(certificate[typ]),
                        "type": "variable",
                    }
        return name, certificate

    def parse_constraints(
        self,
        dico: dict,
    ) -> None:
        if "constraints" not in dico:
            return
        for constraint in dico["constraints"]:
            if "condition" in constraint:
                for condition in constraint["condition"]:
                    self.parse_condition(condition)
            if "check" in constraint:
                for check in constraint["check"]:
                    self.parse_check(check)
            if "fill" in constraint:
                for fill in constraint["fill"]:
                    self.parse_fill(fill)

    def parse_condition(
        self,
        condition: dict,
    ) -> None:
        if "apply_on_fallback" in condition:
            apply_on_fallback = condition.pop("apply_on_fallback")
        else:
            apply_on_fallback = False
        source = self.get_variable_path(condition["source"])
        if not source:
            source = condition["source"]
        name = condition.pop("name")
        prop = name.split("_", 1)[0]
        multi = False
        for target in condition["target"]:
            typ = target.get("type", "variable")
            if typ == "variable":
                variable_path = self.get_variable_path(target["text"])
                if variable_path is None:
                    continue
                variable = self.paths["variable"][variable_path]
                if variable.get("multi", False):
                    multi = True
        if apply_on_fallback:
            condition_value = True
        else:
            if "{{ suffix }}" in source:
                force_param = {"__var": source}
                source = "__var"
            else:
                force_param = None
            condition_value = self.params_condition_to_jinja(
                prop, source, condition["param"], name.endswith("if_in"), multi
            )
            if force_param:
                condition_value.setdefault("params", {}).update(force_param)
        for target in condition["target"]:
            typ = target.get("type", "variable")
            if typ == "variable":
                variable_path = self.get_variable_path(target["text"])
                if variable_path is None:
                    continue
                variable = self.paths["variable"][variable_path]
                variable[prop] = condition_value
            elif typ == "family":
                family_path = self.get_family_path(target["text"])
                if family_path is None:
                    continue
                family = self.paths["family"][family_path]
                family[prop] = condition_value
            elif typ == "iplist":
                list_name = target["text"]
                if list_name in self.lists["ip"]:
                    for ip in self.lists["ip"].pop(list_name):
                        ip[prop] = condition_value
            elif typ == "filelist":
                list_name = target["text"]
                if list_name in self.lists["file"]:
                    for ip in self.lists["file"].pop(list_name):
                        ip[prop] = condition_value
            elif typ == "servicelist":
                list_name = target["text"]
                if list_name in self.lists["service"]:
                    for service in self.lists["service"].pop(list_name):
                        service[prop] = condition_value
            elif typ == "certificatelist":
                list_name = target["text"]
                if list_name in self.lists["certificate"]:
                    for certificat in self.lists["certificate"].pop(list_name):
                        certificat[prop] = condition_value

    def parse_check(
        self,
        check: dict,
    ) -> None:
        for target in check["target"]:
            variable_path = self.get_variable_path(target["text"])
            if variable_path is None:
                continue
            variable = self.paths["variable"][variable_path]
            if "validators" in variable and variable["validators"] is None:
                variable.pop("validators")
            if check.get("type") == "jinja":
                check_value = check["name"]
            else:
                check["param"] = [
                    {"text": variable_path, "type": "variable"}
                ] + check.get("param", [])
                check_value = self.convert_param_function(
                    check, variable.get("multi", False)
                )
            variable.setdefault("validators", []).append(check_value)

    def parse_fill(
        self,
        fill: dict,
    ) -> None:
        for target in fill.pop("target"):
            params = []
            variable_path = self.get_variable_path(target["text"])
            if variable_path in self.paths["dynamic"]:
                variable_path = self.paths["dynamic"][variable_path]
            if variable_path in self.paths["variable"]:
                variable = self.paths["variable"][variable_path]
                if fill.get("type") == "jinja":
                    fill_value = {
                        "type": "jinja",
                        "jinja": fill["name"],
                    }
                else:
                    fill_value = self.convert_param_function(
                        fill, variable.get("multi", False)
                    )
                variable["default"] = fill_value
                if variable.get("mandatory") is False:
                    del variable["mandatory"]
            else:
                raise Exception(
                    f'cannot set fill to unknown variable "{variable_path}"'
                )

    def params_condition_to_jinja(
        self,
        prop: str,
        path: str,
        params: List[dict],
        if_in: bool,
        multi: bool,
    ) -> str:
        new_params = {}
        jinja = "{% if "
        for idx, param in enumerate(params):
            if idx:
                jinja += " or "

            new_param, value = self.get_jinja_param_and_value(param, multi)
            if value:
                jinja += path + " == " + value
            if new_param:
                new_params |= new_param
        if if_in:
            jinja += " %}" + prop + "{% endif %}"
        else:
            jinja += " %}{% else %}" + prop + "{% endif %}"
        ret = {
            "type": "jinja",
            "jinja": jinja,
        }
        if new_params:
            ret["params"] = new_params
        return ret

    def get_value(
        self,
        param: dict,
        multi: bool,
    ) -> Any:
        typ = param.get("type", "string")
        if typ == "string":
            value = param.get("text")
        elif typ == "number":
            value = int(param["text"])
        elif typ == "nil":
            value = None
        elif typ == "space":
            value = " "
        elif typ == "boolean":
            value = param["text"]
        elif typ == "variable":
            variable_path = self.get_variable_path(param["text"])
            if variable_path is None:
                variable_path = "__" + param["text"]
            value = {
                "type": "variable",
                "variable": variable_path,
            }
            if "optional" in param:
                value["optional"] = param["optional"]
            if "propertyerror" in param:
                value["propertyerror"] = param["propertyerror"]
        elif typ == "function":
            value = self.convert_param_function(param, multi)
        elif typ == "information":
            value = {
                "type": "information",
                "information": param["text"],
            }
            if "variable" in param:
                variable_path = self.get_variable_path(param["variable"])
                value["variable"] = variable_path
        elif typ == "suffix":
            value = param
        elif typ == "index":
            value = param
        return value

    def get_jinja_param_and_value(
        self,
        param,
        multi: bool,
    ) -> Tuple[list, Any]:
        new_param = None
        typ = param.get("type", "string")
        value = self.get_value(param, multi)
        if isinstance(value, dict):
            if typ == "information":
                key = normalize_family(value["information"])
                if "variable" in value:
                    attr_name = f'{value["variable"]}.{key}'
                else:
                    attr_name = key
                attr_name = f"__information_{attr_name}"
                new_param = {attr_name: value}
                value = attr_name
            elif typ in ["index", "suffix"]:
                if "name" in value:
                    attr_name = value["name"]
                else:
                    attr_name = f"__{typ}"
                new_param = {attr_name: {"type": typ}}
                value = attr_name
            elif "propertyerror" in param or "optional" in param:
                attr_name = value["variable"]
                new_param = {attr_name: value}
                value = value[typ]
            elif "{{ suffix }}" in value[typ]:
                path = value[typ]
                attr_name = path.split(".")[-1][:-12]  # remove {{ suffix }}
                new_param = {attr_name: value}
                value = attr_name
            else:
                value = value[typ]
            if not value:
                return
        else:
            value = dumps(value, ensure_ascii=False)
        return new_param, value

    def convert_param_function(
        self,
        param: dict,
        multi: bool,
    ) -> str:
        text = param["name"]
        params = {}
        #        multi = False
        if "param" in param and param["param"]:
            if (
                text == "calc_value"
                and len(param["param"]) == 1
                and isinstance(param["param"][0], dict)
                and param["param"][0].get("type") == "variable"
                and param["param"][0].get("text")
            ):
                value = param["param"][0]["text"]
                path = self.get_variable_path(value)
                if not path:
                    path = value
                ret = {"type": "variable", "variable": path}
                if "optional" in param["param"][0]:
                    ret["optional"] = param["param"][0]["optional"]
                return ret
            first, *others = param["param"]
            new_param, first = self.get_jinja_param_and_value(first, multi)
            text = f"{first} | {text}"
            if new_param:
                params |= new_param
            if others:
                values = []
                for param in others:
                    new_param, value = self.get_jinja_param_and_value(param, multi)
                    if new_param:
                        params |= new_param
                    if "name" in param:
                        if param["name"] == "multi" and value == "true":
                            multi = True
                        values.append(f'{param["name"]}={value}')
                    else:
                        values.append(value)
                text += "("
                text += ", ".join(values)
                text += ")"
        else:
            text += "()"
        if not multi:
            text = "{{ " + text + " }}"
        else:
            text = (
                """{% for __variable in """
                + text
                + """ %}
{{ __variable }}
{% endfor %}"""
            )
        ret = {"type": "jinja", "jinja": text}
        if params:
            ret["params"] = params
        return ret

    def get_variable_path(
        self,
        path: str,
    ) -> dict:
        if (
            path not in self.paths["variable"]
            and path in self.flatten_paths["variable"]
        ):
            path = self.flatten_paths["variable"][path]
        if path in self.paths["dynamic"]:
            path = self.paths["dynamic"][path]
        if path not in self.paths["variable"]:
            return path
        return path

    def get_family_path(
        self,
        path: str,
    ) -> dict:
        if path not in self.paths["family"] and path in self.flatten_paths["family"]:
            path = self.flatten_paths["family"][path]
        if path not in self.paths["family"]:
            return
        return path

    def get(self) -> dict:
        return self.variables, self.services


class RougailUpgrade:
    def __init__(
        self,
        test=False,
        upgrade_help=None,
        rougailconfig: RougailConfig = None,
    ) -> None:
        self.test = test
        if upgrade_help is None:
            upgrade_help = {}
        self.upgrade_help = upgrade_help
        if rougailconfig is None:
            rougailconfig = RougailConfig
        self.rougailconfig = rougailconfig

    def run(
        self,
    ):
        for dict_dir, dest_dir in zip(
            self.rougailconfig["main_dictionaries"],
            self.rougailconfig["upgrade_options.main_dictionaries"],
        ):
            self._load_dictionaries(
                dict_dir,
                dest_dir,
                normalize_family(self.rougailconfig["main_namespace"]),
            )
        if self.rougailconfig["main_namespace"]:
            if self.rougailconfig["extra_dictionaries"]:
                dst_extra_dir = self.rougailconfig["upgrade_options.extra_dictionary"]
                for namespace, extra_dirs in self.rougailconfig[
                    "extra_dictionaries"
                ].items():
                    extra_dstsubfolder = Path(dst_extra_dir) / namespace
                    if not extra_dstsubfolder.is_dir():
                        extra_dstsubfolder.mkdir()
                    for extra_dir in extra_dirs:
                        self._load_dictionaries(
                            str(extra_dir),
                            str(extra_dstsubfolder),
                            normalize_family(namespace),
                        )

    def _load_dictionaries(
        self,
        srcfolder: str,
        dstfolder: Optional[str],
        namespace: str,
    ) -> None:
        if dstfolder is None:
            dstfolder = srcfolder
        Path(dstfolder).mkdir(parents=True, exist_ok=True)
        filenames = [
            filename
            for filename in listdir(srcfolder)
            if filename.endswith(".xml") or filename.endswith(".yml")
        ]
        filenames.sort()
        for filename in filenames:
            xmlsrc = Path(srcfolder) / Path(filename)

            ymldst = Path(dstfolder) / (Path(filename).stem + ".yml")
            if filename.endswith(".xml"):
                if parse is None:
                    raise Exception("XML module is not installed")
                try:
                    parser = XMLParser(remove_blank_text=True)
                    document = parse(xmlsrc, parser)
                except XMLSyntaxError as err:
                    raise Exception(_("not a XML file: {0}").format(err)) from err
                root = document.getroot()
                search_function_name = get_function_name(
                    root.attrib.get("version", "1")
                )
                ext = "xml"
            else:
                with xmlsrc.open() as file_fh:
                    root = YAML(typ="safe").load(file_fh)
                search_function_name = get_function_name(str(root["version"]))
                ext = "yml"
            function_found = False
            root_services = None
            for version, function_version in FUNCTION_VERSIONS:
                if function_found and hasattr(self, function_version):
                    upgrade_help = self.upgrade_help.get(function_version, {}).get(
                        filename, {}
                    )
                    if upgrade_help.get("remove") is True:
                        continue
                    root, root_services_, new_type = getattr(self, function_version)(
                        root, upgrade_help, namespace, xmlsrc, ext
                    )
                    if root_services_ is not None:
                        root_services = root_services_
                if function_version == search_function_name:
                    function_found = True
            if root != {"version": None}:
                root["version"] = float(version)
                with ymldst.open("w") as ymlfh:
                    yaml = YAML()
                    yaml.dump(
                        root,
                        ymlfh,
                    )
            # if root_services and services_dstfolder:
            #     root_services["version"] = version
            #     ymldst_services.parent.mkdir(parents=True, exist_ok=True)
            #     with ymldst_services.open("w") as ymlfh:
            #         yaml = YAML()
            #         yaml.dump(
            #             root_services,
            #             ymlfh,
            #         )

    def _attribut_to_bool(self, variable):
        for prop in [
            "mandatory",
            "hidden",
            "redefine",
            "multi",
            "leadership",
            "optional",
            "unique",
            "auto_save",
            "remove_check",
            "manage",
            "exists",
            "disabled",
            "undisable",
            "remove_choice",
            "propertyerror",
            "apply_on_fallback",
            "remove_fill",
            "remove_condition",
        ]:
            if prop in variable:
                variable[prop] = {"True": True, "False": False}[variable[prop]]

    def _attribut_to_int(self, variable):
        for prop in ["mode"]:
            if prop in variable:
                if variable[prop] in ["expert", "normal"]:
                    variable[prop] = {
                        "expert": "advanced",
                        "normal": "standard",
                    }.get(variable[prop])
                    continue
                try:
                    variable[prop] = int(variable[prop])
                except ValueError:
                    pass

    def _xml_to_yaml(
        self,
        objects,
        obj_name,
        variables,
        path,
        variable_type="string",
        variable_choices=[],
    ):
        if obj_name in ["variables", "family"]:
            dico = []
        else:
            dico = {}
        for obj in objects:
            obj_type = obj.tag
            if not isinstance(obj_type, str):
                # doesn't proceed the XML commentaries
                continue
            new_dico = dict(obj.attrib)
            if obj_type in ["variable", "family"]:
                if path:
                    path += "." + obj.attrib["name"]
                else:
                    path = obj.attrib["name"]
            choices = []
            if obj_type == "variable":
                var_type = obj.attrib.get("type", "string")
                variables[obj.attrib["name"]] = {"type": var_type}
                variables[path] = {"type": var_type}
            elif obj_type == "condition":
                var_type = variables.get(obj.attrib["source"], {}).get("type", "string")
                if var_type == "choice":
                    choices = variables.get(obj.attrib["source"], {}).get("choices", [])
            else:
                var_type = None
            new_objects = self._xml_to_yaml(
                obj, obj.tag, variables, path, var_type, choices
            )
            if obj.text:
                text = obj.text
                if isinstance(text, str):
                    text = text.strip()
                if text:
                    if obj_type in ["choice", "value"]:
                        value_type = obj.attrib.get("type")
                        if not value_type and obj_type == "value":
                            value_type = variable_type
                        text = CONVERT_OPTION.get(value_type, {}).get("func", str)(text)
                        if obj_type == "choice":
                            variables[path.rsplit(".", 1)[-1]].setdefault(
                                "choices", []
                            ).append({"type": value_type, "value": text})
                            variables[path].setdefault("choices", []).append(
                                {"type": value_type, "value": text}
                            )
                    if obj_type in ["param", "value"]:
                        if obj.attrib.get("type") == "variable":
                            var_type = variables.get(obj.attrib.get("name"), {}).get(
                                "type", "string"
                            )
                            value_type = obj.attrib.get("type", var_type)
                            text = CONVERT_OPTION.get(value_type, {}).get("func", str)(
                                text
                            )
                        elif "type" in obj.attrib:
                            var_type = obj.attrib.get("type")
                            text = CONVERT_OPTION.get(var_type, {}).get("func", str)(
                                text
                            )
                        elif obj_type == "param" and variable_type:
                            if variable_type == "choice":
                                for choice in variable_choices:
                                    if choice["value"] == CONVERT_OPTION.get(
                                        choice.get("type"), {}
                                    ).get("func", str)(text):
                                        text = choice["value"]
                                        break
                            else:
                                text = CONVERT_OPTION.get(variable_type, {}).get(
                                    "func", str
                                )(text)
                    new_dico["text"] = text
            if isinstance(new_objects, list):
                if new_objects:
                    for new_obj in new_objects:
                        new_dico.update(new_obj)
            elif new_objects is not None and list(new_objects.values())[0]:
                new_dico.update(new_objects)
            self._attribut_to_bool(new_dico)
            self._attribut_to_int(new_dico)
            if not new_dico:
                new_dico = None
            if obj_type == "override" and not new_dico:
                new_dico = None
            if isinstance(dico, list):
                if dico and obj_type in dico[-1]:
                    dico[-1][obj_type].append(new_dico)
                else:
                    dico.append({obj_type: [new_dico]})
            elif new_dico is None:
                dico[obj_type] = new_dico
            else:
                dico.setdefault(obj_type, []).append(new_dico)
        if dico == {}:
            dico = None
        elif isinstance(dico, dict):
            dico = [dico]
        if obj_name in ["service", "condition", "fill", "choice", "check"]:
            pass
        elif obj_name in ["variables", "family"]:
            dico = {"variables": dico}
        elif obj_name != "variable":
            dico = {obj_name: dico}
        return dico

    def _update_1_1(self, root):
        new_root = {}
        update_root = False
        for key, value in root.items():
            new_root[key] = value
            if not isinstance(value, dict):
                continue
            # migrate dynamic family
            if (
                ("variable" in value and isinstance(value["variable"], str))
                or ("_variable" in value and isinstance(value["_variable"], str))
            ) and (
                ("_type" in value and value["_type"] == "dynamic")
                or ("type" in value and value["type"] == "dynamic")
            ):
                value["dynamic"] = {
                    "type": "variable",
                    "variable": value.pop("variable"),
                    "propertyerror": False,
                }
                if "{{ suffix }}" not in key:
                    new_root[key + "{{ suffix }}"] = new_root.pop(key)
                    update_root = True
            self._update_1_1(value)
            for typ, obj in {
                "boolean": bool,
                "number": int,
                "string": str,
                "float": float,
            }.items():
                if value.get("type") == typ:
                    default = value.get("default")
                    if default is None or default == []:
                        continue
                    if isinstance(default, obj):
                        del value["type"]
                    elif isinstance(default, list) and isinstance(default[0], obj):
                        del value["type"]
            if value.get("multi") and isinstance(value.get("default"), list):
                del value["multi"]
        if update_root:
            root.clear()
            root.update(new_root)

    def update_1_1(
        self,
        root,
        upgrade_help: dict,
        namespace: str,
        xmlsrc: str,
        ext: str,
    ):
        self._update_1_1(root)
        return root, None, "yml"

    def update_1_0(
        self,
        root: "Element",
        upgrade_help: dict,
        namespace: str,
        xmlsrc: str,
        ext: str,
    ) -> "Element":
        if ext == "xml":
            new_root = {"version": root.attrib["version"]}
            variables = {}
            for typ in ["services", "variables", "constraints"]:
                objects = root.find(typ)
                if objects is None:
                    objects = []
                new_objects = self._xml_to_yaml(objects, typ, variables, namespace)
                if new_objects[typ]:
                    new_root.update(new_objects)
        else:
            new_root = root
        variables, services = upgrade_010_to_10(new_root, namespace, xmlsrc).get()
        return variables, services, "yml"

    def update_0_10(
        self,
        root: "Element",
        upgrade_help: dict,
        namespace: str,
        xmlsrc: str,
        ext: str,
    ) -> "Element":
        variables = root.find("variables")
        if variables is None:
            return root
        paths = self._get_path_variables(
            variables,
            namespace == "configuration",
            namespace,
        )
        constraints = root.find("constraints")
        # convert schedule and schedulemod
        for variable in paths.values():
            variable = variable["variable"]
            if variable.tag != "variable":
                continue
            if "type" in variable.attrib and variable.attrib["type"] in [
                "schedule",
                "schedulemod",
            ]:
                if variable.attrib["type"] == "schedule":
                    choices = ("none", "daily", "weekly", "monthly")
                else:
                    choices = ("pre", "post")
                variable.attrib["type"] = "choice"
                has_value = False
                for value in variable:
                    if value.tag == "value":
                        has_value = True
                        break
                for name in choices:
                    choice = SubElement(variable, "choice")
                    choice.text = name
                if not has_value:
                    value = SubElement(variable, "value")
                    value.text = choices[0]

        # convert group to leadership
        groups = []
        if constraints is not None:
            for constraint in constraints:
                if constraint.tag == "group":
                    constraints.remove(constraint)
                    groups.append(constraint)
        for group in groups:
            if group.attrib["leader"] in paths:
                leader_obj = paths[group.attrib["leader"]]
                # FIXME name peut avoir "." il faut le virer
                # FIXME si extra c'est un follower !
                if "name" in group.attrib:
                    grpname = group.attrib["name"]
                    if "description" in group.attrib:
                        description = group.attrib["description"]
                    else:
                        description = grpname
                else:
                    grpname = leader_obj["variable"].attrib["name"]
                    if "." in grpname:
                        grpname = grpname.rsplit(".", 1)[-1]
                    if "description" in group.attrib:
                        description = group.attrib["description"]
                    elif "description" in leader_obj["variable"].attrib:
                        description = leader_obj["variable"].attrib["description"]
                    else:
                        description = grpname
                family = SubElement(
                    leader_obj["parent"],
                    "family",
                    name=grpname,
                    description=description,
                    leadership="True",
                )
                leader_obj["parent"].remove(leader_obj["variable"])
                family.append(leader_obj["variable"])
            else:
                # append in group
                follower = next(iter(group))
                leader_name = group.attrib["leader"]
                if "." in leader_name:
                    leader_path = leader_name.rsplit(".", 1)[0]
                    follower_path = leader_path + "." + follower.text
                else:
                    follower_path = follower.text
                obj = paths[follower_path]
                family = SubElement(
                    obj["parent"], "family", name=leader_name, leadership="True"
                )
                grpname = leader_name
            for follower in group:
                leader_name = group.attrib["leader"]
                if "." in leader_name:
                    leader_path = leader_name.rsplit(".", 1)[0]
                    follower_path = leader_path + "." + follower.text
                else:
                    follower_path = follower.text
                follower_obj = paths[follower_path]
                follower_obj["parent"].remove(follower_obj["variable"])
                family.append(follower_obj["variable"])
                if "." in follower_path:
                    new_path = (
                        follower_path.rsplit(".", 1)[0]
                        + "."
                        + grpname
                        + "."
                        + follower_path.rsplit(".", 1)[1]
                    )
                    paths[new_path] = paths[follower_path]

        # convert choice option
        valid_enums = []
        if constraints is not None:
            for constraint in constraints:
                if (
                    constraint.tag == "check"
                    and constraint.attrib["name"] == "valid_enum"
                ):
                    constraints.remove(constraint)
                    valid_enums.append(constraint)
        for valid_enum in valid_enums:
            targets = []
            for target in valid_enum:
                if target.tag != "target":
                    continue
                if target.text in paths:
                    # not in paths if it's optional
                    # but not check it
                    targets.append(paths[target.text]["variable"])
            params = []
            function_param = None
            for param in valid_enum:
                if param.tag != "param":
                    continue
                if "type" in param.attrib and param.attrib["type"] == "function":
                    function_param = param.text
                    continue
                params.append(param)
            first_choice = None
            for target in targets:
                if function_param is not None:
                    function = SubElement(
                        target, "choice", type="function", name=function_param
                    )
                for param in params:
                    if function_param is not None:
                        function.append(param)
                    else:
                        choice = SubElement(target, "choice")
                        if first_choice is None and (
                            "type" not in param.attrib
                            or param.attrib["type"] != "variable"
                        ):
                            first_choice = choice
                        choice.text = param.text
                        if "type" not in param.attrib and param.text is None:
                            choice_type = "nil"
                        elif "type" in param.attrib:
                            choice_type = param.attrib["type"]
                        elif "type" in target.attrib:
                            choice_type = target.attrib["type"]
                        else:
                            choice_type = "string"
                        choice.attrib["type"] = choice_type
            has_value = False
            for target in targets:
                if "remove_check" in target.attrib:
                    target.attrib["remove_choice"] = target.attrib["remove_check"]
            for target in targets:
                for value in target:
                    if value.tag == "value":
                        has_value = True
                        if "type" in target.attrib:
                            value.attrib["type"] = target.attrib["type"]
            if first_choice is not None and not has_value:
                value = SubElement(target, "value")
                value.attrib["type"] = first_choice.attrib["type"]
                value.text = first_choice.text
            for target in targets:
                if (
                    "remove_choice" not in target.attrib
                    or target.attrib["remove_choice"] != "True"
                ):
                    target.attrib["type"] = "choice"
        return root, None, "xml"

    def _get_path_variables(self, variables, is_variable_namespace, path, dico=None):
        if dico is None:
            dico = {}
        for variable in variables:
            if not is_variable_namespace and path:
                subpath = path + "."
            else:
                subpath = ""
            if variable.tag not in ["variable", "family"]:
                continue
            subpath += variable.attrib["name"]
            if variable.tag == "family":
                self._get_path_variables(variable, is_variable_namespace, subpath, dico)
            elif variable.tag == "variable":
                dico[subpath] = {"variable": variable, "parent": variables}
        return dico

    @staticmethod
    def move(elt, src, dst, optional=False):
        if src == "text":
            value = elt.text
            elt.text = None
        else:
            if optional and src not in elt.attrib:
                return
            value = elt.attrib[src]
            del elt.attrib[src]
        #
        if dst == "text":
            elt.text = value
        else:
            elt.attrib[dst] = value

    @staticmethod
    def remove(elt, src, optional=False):
        if optional and src not in elt.attrib:
            return
        del elt.attrib[src]

    @staticmethod
    def create_service(services, service_name, service_elt, servicelists, upgrade_help):
        if service_name in service_elt:
            return service_elt[service_name]
        service = SubElement(services, "service")
        service.attrib["name"] = service_name
        if service_name == "unknown":
            service.attrib["manage"] = "False"
        if service_name in upgrade_help.get("services", {}).get("unmanage", []):
            service.attrib["manage"] = "False"
        service_elt[service_name] = service
        if upgrade_help.get("servicelists", {}).get(service_name):
            service.attrib["servicelist"] = upgrade_help.get("servicelists", {}).get(
                service_name
            )
        elif service_name in servicelists:
            service.attrib["servicelist"] = servicelists[service_name]
        return service

    def upgrade_container(
        self, elt, current_service, files, ip, servicelists, upgrade_help
    ):
        if elt.tag == "file":
            self.move(elt, "name", "text")
            self.remove(elt, "del_comment", optional=True)
            elt.attrib["engine"] = "creole_legacy"
            if (
                not "instance_mode" in elt.attrib
                or elt.attrib["instance_mode"] != "when_container"
            ) and elt.text not in upgrade_help.get("files", {}).get("remove", {}):
                if elt.attrib.get("filelist") in upgrade_help.get("services", {}).get(
                    "filelist_service", {}
                ):
                    elt_service = upgrade_help.get("services", {}).get(
                        "filelist_service", {}
                    )[elt.attrib["filelist"]]
                    if elt_service in files:
                        service = elt_service
                    else:
                        service = current_service
                else:
                    service = current_service
                files[service][elt.text] = elt
        elif elt.tag in [
            "host",
            "disknod",
            "fstab",
            "interface",
            "package",
            "service_access",
        ]:
            pass
        elif elt.tag == "service_restriction":
            for restriction in elt:
                if restriction.tag == "ip" and restriction.text != "0.0.0.0":
                    self.remove(restriction, "ip_type", optional=True)
                    self.remove(restriction, "netmask_type", optional=True)
                    if elt.attrib["service"] in upgrade_help.get("services", {}).get(
                        "rename", {}
                    ):
                        elt_service = upgrade_help.get("services", {}).get(
                            "rename", {}
                        )[elt.attrib["service"]]
                    else:
                        elt_service = elt.attrib["service"]
                    if elt_service in ip:
                        service = elt_service
                    else:
                        service = current_service
                    ip[service].append(restriction)
        elif elt.tag == "service":
            new_name = elt.text
            if current_service == "unknown":
                if new_name in files:
                    raise Exception("hu?")
                files[new_name] = files[current_service]
                del files[current_service]
                ip[new_name] = ip[current_service]
                del ip[current_service]
            elif new_name not in files:
                files[new_name] = {}
                ip[new_name] = []
            current_service = new_name
            if "servicelist" in elt.attrib:
                servicelists[current_service] = elt.attrib["servicelist"]
        else:
            raise Exception(f"unknown containers tag {elt.tag}")
        return current_service



from shutil import rmtree  # , copyfile, copytree
from os import getcwd, makedirs
from os.path import isfile, join, isdir
from pytest import fixture, raises
from os import listdir

from rougail import Rougail, RougailConfig
from rougail.error import DictConsistencyError

from ruamel.yaml import YAML

# dico_dirs = 'tests/data'
# test_ok = set()
# for test in listdir(dico_dirs):
#    if isdir(join(dico_dirs, test)):
#        test_ok.add(test)
# excludes = set([])
# test_ok -= excludes

# ORI_DIR = getcwd()

# test_ok = list(test_ok)
# test_ok.sort()
##print(test_ok)

# @fixture(scope="module", params=test_ok)
# def test_dir(request):
#    return request.param

"""
the kinematics are as follows:
- if no version attribute is defined in the yaml file, then the default version attribute of rougailconfig is taken
- if a version attribute is defined in the yaml file, this is the one that is taken
"""


def test_validate_default_version():
    "retrieves the default_dictionary_format_version if no version in the yaml file"

    RougailConfig["dictionaries_dir"] = ["tests/data/dict1"]
    RougailConfig["default_dictionary_format_version"] = "1.1"
    rougail = Rougail()
    config = rougail.get_config()

    filename = "tests/data/dict1/dict.yml"
    with open(filename, encoding="utf8") as file_fh:
        objects = YAML(typ="safe").load(file_fh)
    version = rougail.converted.validate_file_version(objects, filename)
    assert version == RougailConfig["default_dictionary_format_version"]


def test_validate_file_version_from_yml():
    "retrives the yaml file version defined in the yaml file"

    RougailConfig["dictionaries_dir"] = ["tests/data/dict2"]
    RougailConfig["default_dictionary_format_version"] = "1.1"
    rougail = Rougail()
    config = rougail.get_config()

    filename = "tests/data/dict2/dict.yml"
    with open(filename, encoding="utf8") as file_fh:
        objects = YAML(typ="safe").load(file_fh)
    version = rougail.converted.validate_file_version(objects, filename)
    assert version == "1.0"


def test_retrieve_version_from_config():

    RougailConfig["dictionaries_dir"] = ["tests/data/dict2"]
    RougailConfig["default_dictionary_format_version"] = "1.1"
    rougail = Rougail()
    # FIXME replace with rougail.annotator()
    # rougail.converted.annotator()
    rougail.get_config()
    assert rougail.converted.paths._data["rougail.hello"].version == "1.0"


# def test_dictionary(test_dir):
#    assert getcwd() == ORI_DIR
#    test_dir = join(dico_dirs, test_dir)
#    launch_test(test_dir, 'dict')
#    assert getcwd() == ORI_DIR



"""
Config file for Rougail

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from pathlib import Path
from tiramisu import Config
from ruamel.yaml import YAML
from .utils import _, load_modules, normalize_family
from .convert import RougailConvert


RENAMED = {
    "dictionaries_dir": "main_dictionaries",
    "variable_namespace": "main_namespace",
    "functions_file": "functions_files",
}
NOT_IN_TIRAMISU = {
    "custom_types": {},
}
SUBMODULES = None


def get_sub_modules():
    global SUBMODULES
    if SUBMODULES is None:
        SUBMODULES = {}
        for submodule in Path(__file__).parent.iterdir():
            if submodule.name.startswith("_") or not submodule.is_dir():
                continue
            config_file = submodule / "config.py"
            if config_file.is_file():
                SUBMODULES[submodule.name] = load_modules(
                    "rougail." + submodule.name + ".config", str(config_file)
                )
    return SUBMODULES


def get_level(module):
    return module["level"]


class _RougailConfig:
    def __init__(self, backward_compatibility: bool, root, extra_vars: dict):
        self.backward_compatibility = backward_compatibility
        self.root = root
        self.config = Config(
            self.root,
        )
        self.config.property.read_only()
        self.extra_vars = extra_vars
        self.not_in_tiramisu = NOT_IN_TIRAMISU | extra_vars
        for variable, default_value in self.not_in_tiramisu.items():
            if not isinstance(default_value, str):
                default_value = default_value.copy()
            setattr(self, variable, default_value)

    def copy(self):
        rougailconfig = _RougailConfig(
            self.backward_compatibility, self.root, self.extra_vars
        )
        rougailconfig.config.value.importation(self.config.value.exportation())
        rougailconfig.config.property.importation(self.config.property.exportation())
        rougailconfig.config.property.read_only()
        for variable in self.not_in_tiramisu:
            value = getattr(self, variable)
            if not isinstance(value, str):
                value = value.copy()
            setattr(rougailconfig, variable, value)
        return rougailconfig

    def __setitem__(
        self,
        key,
        value,
    ) -> None:
        if key in self.not_in_tiramisu:
            setattr(self, key, value)
        else:
            self.config.property.read_write()
            if key == "export_with_import":
                key = "not_export_with_import"
            key = RENAMED.get(key, key)
            option = self.config.option(key)
            if option.isoptiondescription() and option.isleadership():
                leader = list(value)
                option.leader().value.reset()
                option.leader().value.set(leader)
                follower = option.followers()[0]
                for idx, val in enumerate(value.values()):
                    self.config.option(follower.path(), idx).value.set(val)
            elif key == "not_export_with_import":
                option.value.set(not value)
            else:
                option.value.set(value)
            self.config.property.read_only()

    def __getitem__(
        self,
        key,
    ) -> None:
        if key in self.not_in_tiramisu:
            return getattr(self, key)
        if key == "export_with_import":
            key = "not_export_with_import"
        option = self.config.option(key)
        if option.isoptiondescription() and option.isleadership():
            return self.get_leadership(option)
        ret = self.config.option(key).value.get()
        if key == "not_export_with_import":
            return not ret
        return ret

    def get_leadership(self, option) -> dict:
        leader = None
        followers = []
        for opt, value in option.value.get().items():
            if opt.issymlinkoption():
                continue
            if leader is None:
                leader = value
            else:
                followers.append(value)
        return dict(zip(leader, followers))

    def parse(self, config) -> str:
        for option in config:
            if option.isoptiondescription():
                yield from self.parse(option)
            elif not option.issymlinkoption():
                yield f"{option.path()}: {option.value.get()}"

    def __repr__(self):
        self.config.property.read_write()
        try:
            values = "\n".join(self.parse(self.config))
        except Exception as err:
            values = str(err)
        self.config.property.read_only()
        return values


class FakeRougailConvert(RougailConvert):
    def __init__(
        self,
        add_extra_options: bool,
    ) -> None:
        self.add_extra_options = add_extra_options
        super().__init__({})

    def load_config(self) -> None:
        self.sort_dictionaries_all = False
        self.main_namespace = None
        self.suffix = ""
        self.custom_types = {}
        self.functions_files = []
        self.modes_level = []
        self.extra_annotators = []
        self.base_option_name = "baseoption"
        self.export_with_import = True
        self.internal_functions = []
        self.plugins = ["structural_commandline"]
        self.add_extra_options = self.add_extra_options


def get_rougail_config(
    *,
    backward_compatibility: bool = True,
    add_extra_options: bool = True,
) -> _RougailConfig:
    if backward_compatibility:
        main_namespace_default = "rougail"
    else:
        main_namespace_default = "null"
    rougail_options = f"""default_dictionary_format_version:
  description: Dictionary format version by default, if not specified in dictionary file
  alternative_name: v
  choices:
    - '1.0'
    - '1.1'
  mandatory: false

main_dictionaries:
  description: 'Directories where dictionary files are placed'
  type: unix_filename
  alternative_name: m
  params:
     allow_relative: True
     test_existence: True
     types:
       - directory
  multi: true

sort_dictionaries_all:
  description: Sort dictionaries from differents directories
  negative_description: Sort dictionaries directory by directory
  default: false

main_namespace:
  description: Main namespace name
  default: {main_namespace_default}
  alternative_name: s
  mandatory: false

extra_dictionaries:
  description: Extra namespaces
  type: leadership
  disabled:
    variable: main_namespace
    when: null

  names:
    description: 'Extra namespace name'
    alternative_name: xn
    multi: true
    mandatory: false

  directories:
    description: Directories where extra dictionary files are placed
    alternative_name: xd
    type: unix_filename
    params:
      allow_relative: true
      test_existence: true
      types:
        - directory
    multi: true

upgrade:
  description: Update dictionaries to newest Rougail format version
  negative_description: Do not update dictionaries to newest Rougail format version
  default: false

upgrade_options:
  description: Update informations
  disabled:
    variable: upgrade
    when: false

  main_dictionaries:
    description: 'Directories where dictionary files will be placed'
    default:
      variable: __.main_dictionaries

  extra_dictionary:
    description: 'Directories where extra files will be placed'
    type: unix_filename
    params:
      allow_relative: true
      test_existence: true
      types:
        - directory
    disabled:
      variable: __.main_namespace
      when: null

functions_files:
  description: File with functions
  alternative_name: c
  type: unix_filename
  params:
    allow_relative: true
    test_existence: true
    types:
      - file
  multi: true
  mandatory: false

modes_level:
  description: All modes level available
  multi: true
  mandatory: false
"""
    if backward_compatibility:
        rougail_options += """
  default:
    - basic
    - standard
    - advanced
"""
    rougail_options += """
default_family_mode:
  description: Default mode for a family
  default:
    jinja: |
      {% if modes_level %}
      {{ modes_level[0] }}
      {% endif %}
  disabled:
    jinja: |
      {% if not modes_level %}
      No mode
      {% endif %}
  validators:
    - type: jinja
      jinja: |
        {% if default_family_mode not in modes_level %}
        not in modes_level ({modes_level})
        {% endif %}
  commandline: false

default_variable_mode:
  description: Default mode for a variable
  default:
    jinja: |
      {% if modes_level %}
        {% if modes_level | length == 1 %}
      {{ modes_level[0] }}
        {% else %}
      {{ modes_level[1] }}
        {% endif %}
      {% endif %}
  disabled:
    jinja: |
      {% if not modes_level %}
      No mode
      {% endif %}
  validators:
    - type: jinja
      jinja: |
        {% if default_variable_mode not in modes_level %}
        not in modes_level ({modes_level})
        {% endif %}
  commandline: false

base_option_name:
  description: Option name for the base option
  default: baseoption
  commandline: false

not_export_with_import:
  description: In cache file, do not importation of Tiramisu and other dependencies
  default: false
  commandline: false

tiramisu_cache:
  description: Tiramisu cache filename
  alternative_name: t
  type: unix_filename
  mandatory: false
  params:
    allow_relative: true

internal_functions:
  description: Name of internal functions that we can use as a function
  multi: true
  mandatory: false
  commandline: false

extra_annotators:
  description: Name of extra annotators
  multi: true
  mandatory: false
  commandline: false

plugins:
  description: Name of Rougail plugins
  multi: true
  mandatory: false
  commandline: false

suffix:
  description: Suffix add to generated option name
  default: ''
  mandatory: false
  commandline: false
"""
    processes = {
        "structural": [],
        "output": [],
        "user data": [],
    }
    for module in get_sub_modules().values():
        data = module.get_rougail_config()
        processes[data["process"]].append(data)
    # reorder
    for process in processes:
        processes[process] = list(sorted(processes[process], key=get_level))
    rougail_process = """step:  # Load and exporter steps
  disabled:
    variable: upgrade"""
    for process in processes:
        if processes[process]:
            objects = processes[process]
            rougail_process += """
  {NAME}:
    description: Select for {NAME}
    alternative_name: {NAME[0]}
    choices:
""".format(
                NAME=normalize_family(process),
            )
            for obj in objects:
                rougail_process += f"      - {obj['name']}\n"
            if process == "structural":
                rougail_process += "    commandline: false"
            elif process == "user data":
                rougail_process += """    multi: true
    mandatory: false
"""
                hidden_outputs = [
                    process["name"]
                    for process in processes["output"]
                    if not process.get("allow_user_data", True)
                ]
                if hidden_outputs:
                    rougail_process += """    hidden:
      type: jinja
      jinja: |
"""
                    for hidden_output in hidden_outputs:
                        rougail_process += """        {% if _.output == 'NAME' %}
        Cannot load user data for NAME output
        {% endif %}
""".replace(
                            "NAME", hidden_output
                        )
            elif objects:
                rougail_process += "    default: {DEFAULT}".format(
                    DEFAULT=objects[0]["name"]
                )
        else:
            if process == 'output':
                prop = 'hidden'
            else:
                prop = 'disabled'
            rougail_process += """
  {NAME}:
    description: Select for {NAME}
    mandatory: false
    {PROP}: true
    multi: true
    default: ["You haven't installed \\\"{NAME}\\\" package for rougail"]
    validators:
      - jinja: Please install a rougail-{NAME}-* package.
""".format(
                NAME=normalize_family(process),
                PROP=prop,
            )
    rougail_options += rougail_process
    convert = FakeRougailConvert(add_extra_options)
    convert._init()
    convert.namespace = None
    convert.parse_root_file(
        "rougail.config",
        "",
        "1.1",
        YAML().load(rougail_options),
    )
    extra_vars = {}
    for process in processes:
        for obj in processes[process]:
            if "extra_vars" in obj:
                extra_vars |= obj["extra_vars"]
            if not "options" in obj:
                continue
            convert.parse_root_file(
                f'rougail.config.{obj["name"]}',
                "",
                "1.1",
                YAML().load(obj["options"]),
            )

    tiram_obj = convert.save(None)
    optiondescription = {}
    exec(tiram_obj, {}, optiondescription)  # pylint: disable=W0122
    return _RougailConfig(
        backward_compatibility,
        optiondescription["option_0"],
        extra_vars=extra_vars,
    )


RougailConfig = get_rougail_config()



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['default_1.rougail.var3'] = "{% if _.var1 == 'value' or _.var2 == 'blah' %}\nvalue\n{% endif %}\n"
dict_env['default_2.rougail.var3'] = "{% if _.var1 == 'value' or _.var2 == 'blah' %}\nvalue\n{% endif %}\n"
option_3 = StrOption(name="var1", doc="a first variable", default="value", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_4 = StrOption(name="var2", doc="a second variable", properties=frozenset({"basic", "mandatory", Calculation(func['variable_to_property'], Params((ParamValue("disabled"), ParamOption(option_3)), kwargs={'when': ParamValue("value"), 'inverse': ParamValue(False)}), help_function=func['variable_to_property'])}), informations={'type': 'string'})
option_5 = StrOption(name="var3", doc="a third variable", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_1.rougail.var3"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/04_1default_calculation_hidden/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("1.rougail.var3"), '_.var2': ParamOption(option_4, notraisepropertyerror=True), '_.var1': ParamOption(option_3, notraisepropertyerror=True)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_3, option_4, option_5], properties=frozenset({"basic"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"basic"}))
option_8 = StrOption(name="var1", doc="a first variable", default="value", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_9 = StrOption(name="var2", doc="a second variable", properties=frozenset({"basic", "mandatory", Calculation(func['variable_to_property'], Params((ParamValue("disabled"), ParamOption(option_8)), kwargs={'when': ParamValue("value"), 'inverse': ParamValue(False)}), help_function=func['variable_to_property'])}), informations={'type': 'string'})
option_10 = StrOption(name="var3", doc="a third variable", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_2.rougail.var3"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/04_1default_calculation_hidden/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("2.rougail.var3"), '_.var2': ParamOption(option_9, notraisepropertyerror=True), '_.var1': ParamOption(option_8, notraisepropertyerror=True)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_7 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_8, option_9, option_10], properties=frozenset({"basic"}))
optiondescription_6 = OptionDescription(name="2", doc="2", children=[optiondescription_7], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_6])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
option_2 = StrOption(name="var", doc="a variable", properties=frozenset({"force_default_on_freeze", "frozen", "standard"}), informations={'type': 'string'})
optiondescription_1 = OptionDescription(name="family", doc="a family", children=[option_2], properties=frozenset({"hidden", "standard"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1])



"""Annotate to add specify attribute for tiramisu-cmdline

Silique (https://www.silique.fr)
Copyright (C) 2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from rougail.annotator.variable import Walk
from rougail.utils import _
from rougail.error import DictConsistencyError


class Annotator(Walk):
    """Annotate value"""

    level = 80

    def __init__(self, objectspace, *args) -> None:
        if not objectspace.paths:
            return
        self.alternative_names = {}
        self.objectspace = objectspace
        not_for_commandlines = []
        for family in self.get_families():
            if family.commandline:
                continue
            self.not_for_commandline(family)
            not_for_commandlines.append(family.path + ".")
        for variable in self.get_variables():
            if variable.type == "symlink":
                continue
            variable_path = variable.path
            for family_path in not_for_commandlines:
                if variable_path.startswith(family_path):
                    break
            else:
                if not variable.commandline:
                    self.not_for_commandline(variable)
                else:
                    self.manage_alternative_name(variable)
                    self.manage_negative_description(variable)

    def not_for_commandline(self, variable) -> None:
        self.objectspace.properties.add(variable.path, "not_for_commandline", True)

    def manage_alternative_name(self, variable) -> None:
        if not variable.alternative_name:
            return
        alternative_name = variable.alternative_name
        variable_path = variable.path
        all_letters = ""
        for letter in alternative_name:
            all_letters += letter
            if all_letters == "h":
                msg = _('alternative_name "{0}" conflict with "--help"').format(
                    alternative_name
                )
                raise DictConsistencyError(msg, 202, variable.xmlfiles)
            if all_letters in self.alternative_names:
                msg = _('conflict alternative_name "{0}": "{1}" and "{2}"').format(
                    alternative_name, variable_path, self.alternative_names[all_letters]
                )
                raise DictConsistencyError(msg, 202, variable.xmlfiles)

        self.alternative_names[alternative_name] = variable_path
        if "." not in variable_path:
            path = alternative_name
        else:
            path = variable_path.rsplit(".", 1)[0] + "." + alternative_name
        self.objectspace.add_variable(
            alternative_name,
            {"type": "symlink", "path": path, "opt": variable},
            variable.xmlfiles,
            False,
            False,
            variable.version,
        )

    def manage_negative_description(self, variable) -> None:
        if not variable.negative_description:
            if variable.type == "boolean" and not self.objectspace.add_extra_options:
                raise DictConsistencyError(
                    _(
                        'negative_description is mandatory for boolean variable, but "{0}" hasn\'t'
                    ).format(variable.path),
                    200,
                    variable.xmlfiles,
                )
            return
        if variable.type != "boolean":
            raise DictConsistencyError(
                _(
                    'negative_description is only available for boolean variable, but "{0}" is "{1}"'
                ).format(variable.path, variable.type),
                201,
                variable.xmlfiles,
            )
        self.objectspace.informations.add(
            variable.path, "negative_description", variable.negative_description
        )



"""
Silique (https://www.silique.fr)
Copyright (C) 2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from .update import RougailUpgrade



from importlib.machinery import SourceFileLoader as _SourceFileLoader
from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec
class func:
    pass

def _load_functions(path):
    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func
    loader = _SourceFileLoader('func', path)
    spec = _spec_from_loader(loader.name, loader)
    func_ = _module_from_spec(spec)
    loader.exec_module(func_)
    for function in dir(func_):
        if function.startswith('_'):
            continue
        setattr(func, function, getattr(func_, function))
_load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    from tiramisu4 import *
    from tiramisu4.setting import ALLOWED_LEADER_PROPERTIES
except:
    from tiramisu import *
    from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("normal")
ALLOWED_LEADER_PROPERTIES.add("expert")
option_2 = StrOption(name="mode_conteneur_actif", doc="No change", default="oui", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_1 = OptionDescription(name="general", doc="général", children=[option_2], properties=frozenset({"normal"}))
optiondescription_28 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_1], properties=frozenset({"normal"}))
option_9 = FilenameOption(name="name", doc="name", default="/etc/mailname")
option_10 = StrOption(name="source", doc="source", default="mailname")
option_8 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_7 = OptionDescription(name="mailname", doc="mailname", children=[option_9, option_10, option_8])
option_13 = FilenameOption(name="name", doc="name", default="/rougail.conf")
option_14 = StrOption(name="source", doc="source", default="rougail.conf")
option_12 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_11 = OptionDescription(name="rougail_conf", doc="rougail.conf", children=[option_13, option_14, option_12])
optiondescription_6 = OptionDescription(name="files", doc="files", children=[optiondescription_7, optiondescription_11])
option_5 = BoolOption(name="activate", doc="activate", default=True)
option_15 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_30 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_6, option_5, option_15])
optiondescription_30.impl_set_information('type', "service")
optiondescription_29 = OptionDescription(name="services", doc="services", children=[optiondescription_30], properties=frozenset({"hidden"}))
optiondescription_27 = OptionDescription(name="1", doc="1", children=[optiondescription_28, optiondescription_29])
option_4 = StrOption(name="mode_conteneur_actif", doc="No change", default="oui", properties=frozenset({"force_default_on_freeze", "frozen", "hidden", "mandatory", "normal"}))
optiondescription_3 = OptionDescription(name="general", doc="général", children=[option_4], properties=frozenset({"normal"}))
optiondescription_32 = OptionDescription(name="rougail", doc="Rougail", children=[optiondescription_3], properties=frozenset({"normal"}))
option_20 = FilenameOption(name="name", doc="name", default="/etc/mailname")
option_21 = StrOption(name="source", doc="source", default="mailname")
option_19 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_18 = OptionDescription(name="mailname", doc="mailname", children=[option_20, option_21, option_19])
option_24 = FilenameOption(name="name", doc="name", default="/rougail.conf")
option_25 = StrOption(name="source", doc="source", default="rougail.conf")
option_23 = BoolOption(name="activate", doc="activate", default=True)
optiondescription_22 = OptionDescription(name="rougail_conf", doc="rougail.conf", children=[option_24, option_25, option_23])
optiondescription_17 = OptionDescription(name="files", doc="files", children=[optiondescription_18, optiondescription_22])
option_16 = BoolOption(name="activate", doc="activate", default=True)
option_26 = BoolOption(name="manage", doc="manage", default=True)
optiondescription_34 = OptionDescription(name="test_service", doc="test.service", children=[optiondescription_17, option_16, option_26])
optiondescription_34.impl_set_information('type', "service")
optiondescription_33 = OptionDescription(name="services", doc="services", children=[optiondescription_34], properties=frozenset({"hidden"}))
optiondescription_31 = OptionDescription(name="2", doc="2", children=[optiondescription_32, optiondescription_33])
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_27, optiondescription_31])



from tiramisu import *
from tiramisu.setting import ALLOWED_LEADER_PROPERTIES
from re import compile as re_compile
from rougail.tiramisu import func, dict_env, load_functions, ConvertDynOptionDescription
load_functions('tests/dictionaries/../eosfunc/test.py')
try:
    groups.namespace
except:
    groups.addgroup('namespace')
ALLOWED_LEADER_PROPERTIES.add("basic")
ALLOWED_LEADER_PROPERTIES.add("standard")
ALLOWED_LEADER_PROPERTIES.add("advanced")
dict_env['default_1.rougail.var2'] = "{{ _.dynval1.family.var }}\n"
dict_env['default_2.rougail.var2'] = "{{ _.dynval1.family.var }}\n"
option_3 = StrOption(name="var1", doc="a suffix variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_6 = StrOption(name="var", doc="with a variable", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_5 = OptionDescription(name="family", doc="a family", children=[option_6], properties=frozenset({"basic"}))
optiondescription_4 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="a dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_3)))), children=[optiondescription_5], properties=frozenset({"basic"}), informations={'dynamic_variable': '1.rougail.var1'})
option_7 = StrOption(name="var2", doc="a second variable", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_1.rougail.var2"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_2family_dynamic_jinja_fill_sub_group/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("1.rougail.var2"), '_.dynval1.family.var': ParamDynOption(option_6, ["val1"], notraisepropertyerror=True)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_2 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_3, optiondescription_4, option_7], properties=frozenset({"basic"}))
optiondescription_1 = OptionDescription(name="1", doc="1", children=[optiondescription_2], properties=frozenset({"basic"}))
option_10 = StrOption(name="var1", doc="a suffix variable", multi=True, default=["val1", "val2"], default_multi="val1", properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
option_13 = StrOption(name="var", doc="with a variable", properties=frozenset({"basic", "mandatory"}), informations={'type': 'string'})
optiondescription_12 = OptionDescription(name="family", doc="a family", children=[option_13], properties=frozenset({"basic"}))
optiondescription_11 = ConvertDynOptionDescription(name="dyn{{ identifier }}", doc="a dynamic family", identifiers=Calculation(func['calc_value'], Params((ParamOption(option_10)))), children=[optiondescription_12], properties=frozenset({"basic"}), informations={'dynamic_variable': '2.rougail.var1'})
option_14 = StrOption(name="var2", doc="a second variable", default=Calculation(func['jinja_to_function'], Params((), kwargs={'__internal_jinja': ParamValue("default_2.rougail.var2"), '__internal_type': ParamValue("string"), '__internal_multi': ParamValue(False), '__internal_files': ParamValue(['tests/dictionaries/60_2family_dynamic_jinja_fill_sub_group/dictionaries/rougail/00-base.yml']), '__internal_attribute': ParamValue("default"), '__internal_variable': ParamValue("2.rougail.var2"), '_.dynval1.family.var': ParamDynOption(option_13, ["val1"], notraisepropertyerror=True)})), properties=frozenset({"mandatory", "standard"}), informations={'type': 'string'})
optiondescription_9 = OptionDescription(name="rougail", doc="Rougail", group_type=groups.namespace, children=[option_10, optiondescription_11, option_14], properties=frozenset({"basic"}))
optiondescription_8 = OptionDescription(name="2", doc="2", children=[optiondescription_9], properties=frozenset({"basic"}))
option_0 = OptionDescription(name="baseoption", doc="baseoption", children=[optiondescription_1, optiondescription_8])



"""Standard error classes

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022-2024

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from .i18n import _


def display_xmlfiles(xmlfiles: list) -> str:
    """The function format xmlfiles informations to generate errors"""
    if len(xmlfiles) == 1:
        return '"' + xmlfiles[0] + '"'
    return '"' + '", "'.join(xmlfiles[:-1]) + '"' + " and " + '"' + xmlfiles[-1] + '"'


class ConfigError(Exception):
    """Standard error for templating"""


class FileNotFound(ConfigError):
    """Template file is not found"""


class TemplateError(ConfigError):
    """Templating generate an error"""


class TemplateDisabled(TemplateError):
    """Template is disabled."""


class SpaceObjShallNotBeUpdated(Exception):
    """Specific behavior in case of the presence or not
    of an object in the space object
    """


class DictConsistencyError(Exception):
    """It's not only that the Creole XML is valid against the Creole DTD
    it's that it is not consistent.
    """

    def __init__(self, msg, errno, xmlfiles):
        if xmlfiles:
            msg = _("{0} in {1}").format(msg, display_xmlfiles(xmlfiles))
        super().__init__(msg)
        self.errno = errno


class UpgradeError(Exception):
    """Error during XML upgrade"""


## ---- generic exceptions ----


class NotFoundError(Exception):
    "not found error"
    pass


## ---- specific exceptions ----


class VariableCalculationDependencyError(Exception):
    pass
